# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
from qtpy import QtCore
from glue.app.qt import GlueApplication


__all__ = ['toggle_viewer', 'select_viewer', 'create_glue_app',
           'enter_slice_text', 'enter_wavelength_text', 'reset_app_state',
           'sync_all_viewers', 'assert_all_viewer_indices',
           'assert_slice_text']


TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')


def left_click(qtbot, widget):
    qtbot.mouseClick(widget, QtCore.Qt.LeftButton)

def left_button_press(qtbot, layout):
    widget = layout._slice_controller._slice_slider
    qtbot.keyPress(widget, QtCore.Qt.Key_Left)

def right_button_press(qtbot, layout):
    widget = layout._slice_controller._slice_slider
    qtbot.keyPress(widget, QtCore.Qt.Key_Right)

def toggle_viewer(qtbot, layout):
    left_click(qtbot, layout.button_toggle_image_mode)

def select_viewer(qtbot, viewer):
    left_click(qtbot, viewer._widget)

def enter_slice_text(qtbot, layout, text):
    widget = layout._slice_controller._slice_textbox
    widget.setText(str(text))
    qtbot.keyClick(widget, QtCore.Qt.Key_Enter)

def enter_wavelength_text(qtbot, layout, text):
    widget = layout._slice_controller._wavelength_textbox
    widget.setText(str(text))
    qtbot.keyClick(widget, QtCore.Qt.Key_Enter)

def sync_all_viewers(qtbot, layout):
    left_click(qtbot, layout.ui.sync_button)

def assert_viewer_indices(viewer_array, index):
    for viewer in viewer_array:
        assert viewer._widget.slice_index == index

def assert_all_viewer_indices(layout, index):
    assert_viewer_indices(layout.cube_views, index)

def assert_wavelength_text(layout, text):
    current_text = layout._slice_controller._wavelength_textbox.text()
    assert current_text == str(text), "expected={}, current={}".format(text, current_text)

def assert_slice_text(layout, text):
    assert layout._slice_controller._slice_textbox.text() == str(text)


def create_glue_app():
    filename = os.path.join(TEST_DATA_PATH, 'data_cube.fits.gz')

    # We need to make sure that the data factories have been instantiated
    # before creating the glue application below. Otherwise the test data file
    # will not be recognized and the application will hang waiting for user input.
    from ..data_factories import DataFactoryConfiguration

    # We don't want the IFU checker dialog to popup during testing
    # so set check_ifu_valid to False
    dfc = DataFactoryConfiguration(check_ifu_valid=False)

    app = GlueApplication()
    app.run_startup_action('cubeviz')
    app.load_data(filename)
    app.setVisible(True)

    return app

def reset_app_state(qtbot, layout):
    sync_all_viewers(qtbot, layout)
    # Restore the text and index to a known state
    enter_slice_text(qtbot, layout, '1024')
    if layout._single_viewer_mode:
        toggle_viewer(qtbot, layout)
    if layout._active_view is not layout.split_views[0]:
        select_viewer(qtbot, layout.split_views[0])



import os
import yaml
import warnings

import numpy as np

from qtpy.QtCore import Qt, Signal, QThread
from qtpy.QtWidgets import (
    QDialog, QApplication, QPushButton, QProgressBar,
    QLabel, QWidget, QDockWidget, QHBoxLayout, QVBoxLayout,
    QComboBox, QMessageBox, QLineEdit, QRadioButton, QWidgetItem, QSpacerItem
)

from glue.utils.qt import update_combobox

from astropy.utils.exceptions import AstropyWarning
from astropy import units as u
from astropy.wcs.utils import proj_plane_pixel_area
from specviz.third_party.glue.data_viewer import dispatch as specviz_dispatch

from ..messages import FluxUnitsUpdateMessage

OBS_WAVELENGTH_TEXT = 'Obs Wavelength'
REST_WAVELENGTH_TEXT = 'Rest Wavelength'

DEFAULT_FLUX_UNITS_CONFIGS = os.path.join(os.path.dirname(__file__), 'registered_flux_units.yaml')


def _add_unit_to_list(unit_list, target_unit):
    """
    Given a list of units, add target units
    after checking for duplicates
    :param unit_list: list of units or string
    :param target_unit: unit or string
    :return: updated unit
    """

    if isinstance(target_unit, str):
        current_unit = u.Unit(target_unit)
        string_unit = target_unit
    else:
        current_unit = target_unit
        string_unit = target_unit.to_string()

    duplicate = False
    for unit in unit_list:
        if isinstance(unit, str):
            unit = u.Unit(unit)
        if unit == current_unit:
            duplicate = True
            break
    if not duplicate:
        unit_list.append(string_unit)
    return unit_list


def find_unit_index(unit_list, target_unit):
    """
    Given a list of units or strings, find
    index of unit.
    :param unit_list: list of units or string
    :param target_unit: unit or string
    :return: updated unit
    """
    if isinstance(target_unit, str):
        target_unit = u.Unit(target_unit)

    for i, unit in enumerate(unit_list):
        if isinstance(unit, str):
            unit = u.Unit(unit)
        if unit == target_unit:
            return i
    return None


def _get_power(numeric):
    """
    Given a number, get the power if
    the power is an int. Else Return 0
    :param numeric: float or int
    :return: power if int, 0.0 if power is float
    """
    if float(np.log10(numeric)).is_integer():
        power = int(np.log10(numeric))
    else:
        power = 0
    return power


class FluxUnitRegistry:
    """
    Saves a list of spectral flux density units.
    """
    def __init__(self):
        self._model_unit = u.Jy  # Will be used for compatibility checks
        self.runtime_defined_units = []  # Stores list of runtime units

    @staticmethod
    def _locally_defined_units():
        """
        list of defined spectral flux density units.
        :return: list of unit strings
        """
        units = ['Jy', 'mJy', 'uJy',
                 'W / (m2 Hz)',
                 'eV / (s m2 Hz)',
                 'erg / (s cm2)',
                 'erg / (s cm2 um)',
                 'erg / (s cm2 Angstrom)',
                 'erg / (s cm2 Hz)',
                 'ph / (s cm2 um)',
                 'ph / (s cm2 Angstrom)',
                 'ph / (s cm2 Hz)']
        return units

    def is_compatible(self, unit):
        """
        Check if unit can be converted to units
        in this registry. Uses model unit to check.
        :param unit: Unit or unit str
        :return: bool
        """
        try:
            self._model_unit.to(unit, equivalencies=u.spectral_density(3500 * u.AA))
            return True
        except u.UnitConversionError:
            return False

    def get_unit_list(self, current_unit=None):
        """
        Returns a list of unit strings in the registry.
        Adds current_unit if not duplicated. Registered
        units are added first.
        :param current_unit: Unit or unit str
        :return: list of unit str
        """
        unit_list = []  # final list
        item_list = []  # items to be checked if duplicated
        unit_list.extend(self._locally_defined_units())
        item_list.extend(self.runtime_defined_units)

        for item in item_list:
            unit_list = _add_unit_to_list(unit_list, item)

        if current_unit is not None:
            if isinstance(current_unit, str):
                current_unit = u.Unit(current_unit)
            unit_list = _add_unit_to_list(unit_list, current_unit)
        return unit_list

    def add_unit(self, item):
        """
        Add new unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_defined_units:
                self.runtime_defined_units.append(item)


class AreaUnitRegistry:
    """
    Saves a list of area units.
    There are 2 kinds of area units:
    1. solid angles
    2. pixels
    These maybe interchangeable if wcs pixel_scale
    is available.
    """
    def __init__(self):
        self._model_unit = [u.pixel, u.steradian]  # Will be used for compatibility checks
        self.runtime_solid_angle_units = []  # Stores list of runtime angle units
        self.runtime_pixel_units = []  # Stores list of runtime pixel units

    @staticmethod
    def _locally_defined_solid_angle_units():
        """
        list of defined spectral solid angle units.
        :return: list of unit strings
        """
        units = ["deg2",
                 "arcmin2",
                 "arcsec2",
                 'steradian',
                 "rad2"]

        return units

    @staticmethod
    def _locally_defined_pixel_units():
        """
        list of defined spectral pixel units.
        :return: list of unit strings
        """
        units = ["pixel"]

        spaxel = u.def_unit('spaxel', u.astrophys.pix)
        u.add_enabled_units(spaxel)
        units.append('spaxel')

        return units

    def is_compatible(self, unit):
        """
        Check if unit can be converted to units
        in this registry. Uses model units to check.
        :param unit: Unit or unit str
        :return: bool
        """
        compatible = False
        for model_unit in self._model_unit:
            try:
                model_unit.to(unit)
                compatible = True
            except u.UnitConversionError:
                continue
        return compatible

    def get_unit_list(self, pixel_only=False,
                      solid_angle_only=False,
                      current_unit=None):
        """
        Returns a list of unit strings in the registry.
        Adds current_unit if not duplicated. Registered
        units are added first.
        :param pixel_only: (bool) Return pixel units only
        :param solid_angle_only: (bool) Return solid_angle_only units only
        :param current_unit: Unit or unit str
        :return: list of unit str
        """
        unit_list = []
        item_list = []
        if not solid_angle_only:
            unit_list.extend(self._locally_defined_pixel_units())
            item_list.extend(self.runtime_pixel_units)
        if not pixel_only:
            unit_list.extend(self._locally_defined_solid_angle_units())
            item_list.extend(self.runtime_solid_angle_units)

        for item in item_list:
            unit_list = _add_unit_to_list(unit_list, item)

        if current_unit is not None:
            if isinstance(current_unit, str):
                current_unit = u.Unit(current_unit)
            unit_list = _add_unit_to_list(unit_list, current_unit)
        return unit_list

    def add_pixel_unit(self, item):
        """
        Add new pixel unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_pixel_units:
                self.runtime_pixel_units.append(item)

    def add_solid_angle_unit(self, item):
        """
        Add new solid angle unit
        :param item: unit or unit str
        """
        if isinstance(item, str) \
                or isinstance(item, u.UnitBase):
            if item not in self.runtime_solid_angle_units:
                self.runtime_solid_angle_units.append(item)

    def add_unit(self, item):
        """
        Check for type and add new unit
        :param item: unit or unit str
        """
        if isinstance(item, str):
            new_unit = u.unit(item)
        else:
            new_unit = item

        if new_unit.decompose() == u.pix.decompose():
            self.add_pixel_unit(new_unit)
        if 'solid angle' in new_unit.physical_type:
            self.add_solid_angle_unit(new_unit)


FLUX_UNIT_REGISTRY = FluxUnitRegistry()
AREA_UNIT_REGISTRY = AreaUnitRegistry()


class CubeVizUnit:
    """
    Unit container with the following functions:
        - Stores units and names
        - Converts units
        - Updates units
        - Populates conversion gui with widgets custom to
          the stored unit type

    Types of CubeVizUnit:
        - CubeVizUnit (base class)
        - SpectralFluxDensity
            - FormattedSpectralFluxDensity
            - SpectralFluxDensity
        - UnknownUnit
        - NoneUnit
    """
    def __init__(self, unit=None, unit_string=""):
        """
        :param unit: astropy unit or None
        :param unit_string: Unit as a string
        """
        self._controller = None  # Unit controller (property)
        self._original_unit = unit  # the data's actual units
        self._original_unit_string = unit_string  # original_unit as str
        self._unit = unit  # Current Unit
        self._unit_string = unit_string  # unit as str
        self._type = "CubeVizUnit"  # Type of CubeVizUnit
        self.is_convertible = False  # If unit is convertible

        self.message_box = None  # Pointer to gui message box

    @property
    def unit(self):
        return self._unit

    @property
    def unit_string(self):
        return self._unit_string

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        if isinstance(type, str):
            self._type = type

    @property
    def controller(self):
        return self._controller

    @controller.setter
    def controller(self, controller):
        self._controller = controller

    def get_units(self):
        return self.unit_string

    def convert_from_original_unit(self, value, **kwargs):
        """
        Given a value from the data, convert it to current
        units.
        :param value: float
        :param kwargs:
        :return: converted value
        """
        if not isinstance(value, int) \
                and not isinstance(value, float) \
                and not isinstance(value, np.ndarray):
            raise ValueError("Expected float or int, got {} instead.".format(type(value)))

        if self.unit is None:
            return value

        new_value = value
        new_value *= self._original_unit.to(self._unit)

        if isinstance(new_value, u.Quantity):
            new_value = new_value.value

        return new_value

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """
        return True

    def reset_widgets(self):
        """
        Reset all attributes pointing to widgets in the
        conversion gui
        """
        self.message_box = None
        return

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets.
        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        if self.unit_string:
            default_message = "CubeViz can not convert this unit: {0}."
            default_message = default_message.format(self.unit_string)
        else:
            default_message = "Unit not found."

        default_label = QLabel(default_message)
        unit_layout.addWidget(default_label)
        return unit_layout

    def set_message_box(self, message_box):
        """
        Sets local attribute to message box on
        the onversion gui.
        :param message_box: QLabel
        """
        self.message_box = message_box
        self.message_box.setText("")


class SpectralFluxDensity(CubeVizUnit):
    """
    CubeVizUnit for spectral flux density.
    See CubeVizUnit for more info.
    SpectralFluxDensity breaks the units into:
        - power
        - spectral_flux_density
        - area
    These units are recombined as follows:

        unit = (10 ** power) * spectral_flux_density / area

    """
    def __init__(self, unit, unit_string,
                 power=None,
                 spectral_flux_density=None,
                 area=None,
                 is_formatted=False):
        """
        :param unit: astropy unit
        :param unit_string: unit as string
        :param power: power of scalar multiplier if multiplier is 10^X
        :param spectral_flux_density: astropy unit of spectral flux density
        :param area: astropy unit area
        :param is_formatted: True if unit is in registered yaml file
        """
        super(SpectralFluxDensity, self).__init__(unit, unit_string)

        self._type = "FormattedSpectralFluxDensity" if is_formatted else "SpectralFluxDensity"
        self.is_convertible = True

        self.power = power
        self.spectral_flux_density = spectral_flux_density
        self.area = area

        self._original_power = power
        self._original_spectral_flux_density = spectral_flux_density
        self._original_area = area

        self.has_area = True if area is not None else False

        self.power_input = None
        self.flux_combo = None
        self.area_combo = None

        self.wave = 656.3 * u.nm  # Used for conversion preview. Updated latter.

        # Add unit to FLUX_UNIT_REGISTRY. (not added if duplicate)
        FLUX_UNIT_REGISTRY.add_unit(spectral_flux_density)

    def convert_from_original_unit(self, value, wave=None, **kwargs):
        """
        Given a value from the data, convert it to current
        units.
        :param value: float
        :param wave: float: wavelength
        :param kwargs:
        :return: converted value
        """
        if self.unit is None and wave is None:
            return value

        new_value = value

        new_value *= 10**(self._original_power - self.power)
        new_value *= self._original_spectral_flux_density.to(self.spectral_flux_density,
                                                             equivalencies=u.spectral_density(wave))
        if self.has_area:
            pixel_area = self.controller.pixel_area

            if self.area.decompose() == u.pix.decompose() \
                    and 'solid angle' in self._original_area.physical_type:
                area = u.Unit((self._original_area / pixel_area).decompose())
                new_value /= area.to(self.area)
            elif 'solid angle' in self.area.physical_type \
                    and self._original_area.decompose() == u.pix.decompose():
                area = u.Unit((self._original_area * pixel_area).decompose())
                new_value /= area.to(self.area)
            else:
                new_value /= self._original_area.to(self.area)

        if isinstance(new_value, u.Quantity):
            new_value = new_value.value

        return new_value

    def reset_widgets(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """
        self.message_box = None
        self.power_input = None
        self.flux_combo = None
        self.area_combo = None

    def _get_current_power(self):
        """
        :return: current power, 0 if none
        """
        if self.power is not None:
            power = self.power
        else:
            power = 0
        return power

    def _validate_input(self):
        """
        Validate user input from converter gui.
        :return: bool: True if input is valid
        """
        red = "background-color: rgba(255, 0, 0, 128);"
        success = True
        if self.power_input is None \
                or self.flux_combo is None:
            success = False
            return success

        if self.area_combo is None and self.has_area:
            success = False
            return success

        # Check 1: power_input
        if self.power_input == "":
            self.power_input.setStyleSheet(red)
            success = False
        else:
            try:
                power = int(self.power_input.text())
                self.power_input.setStyleSheet("")
            except ValueError:
                self.power_input.setStyleSheet(red)
                success = False
        return success

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """

        success = self._validate_input()
        if not success:
            return success

        self.power = int(self.power_input.text())

        flux_string = self.flux_combo.currentText()
        flux_unit = u.Unit(flux_string)
        self.spectral_flux_density = flux_unit

        if self.has_area:
            area_string = self.area_combo.currentText()
            area_unit = u.Unit(area_string)
            self.area = area_unit
            unit_base = self.spectral_flux_density / self.area
        else:
            unit_base = self.spectral_flux_density

        self._unit = u.Unit((10**self.power) * unit_base)
        self._unit_string = self._unit.to_string()

        return success

    def _update_message(self):
        """
        Callback for when options in conversion gui change. Message
        is set to info about the units a preview of the conversion.
        """
        if self.message_box is None:
            return
        success = self._validate_input()
        if not success:
            self.message_box.setText("Error")
            return

        new_value = 1.0
        wave = self.wave

        power = int(self.power_input.text())
        new_value *= 10**(self._original_power - power)

        flux_string = self.flux_combo.currentText()
        flux_unit = u.Unit(flux_string)
        spectral_flux_density = flux_unit
        new_value *= self._original_spectral_flux_density.to(spectral_flux_density,
                                                             equivalencies=u.spectral_density(wave))

        pixel_area = None
        if self.has_area:
            area_string = self.area_combo.currentText()
            area = u.Unit(area_string)
            pixel_area = self.controller.pixel_area
            if area.decompose() == u.pix.decompose() \
                    and 'solid angle' in self._original_area.physical_type:
                temp_area = (self._original_area / pixel_area).decompose()
                new_value /= temp_area.to(area)
            elif 'solid angle' in area.physical_type \
                    and self._original_area.decompose() == u.pix.decompose():
                temp_area = (self._original_area * pixel_area).decompose()
                new_value /= temp_area.to(area)
            else:
                new_value /= self._original_area.to(area)
            unit_base = spectral_flux_density / area
        else:
            unit_base = spectral_flux_density

        unit_base = u.Unit((10**power) * unit_base)

        if isinstance(new_value, u.Quantity):
            new_value = new_value.value

        if isinstance(unit_base, u.Quantity):
            unit_base = u.Unit(unit_base)

        if pixel_area is None:
            message_param = (wave, self._original_unit.to_string(), new_value, unit_base)
            if 0.01 <= abs(new_value) <= 1000:
                message = "Original Units: [{1}]\n\n" \
                          "New Units: [{3}]\n\n" \
                          "1.0 [Original Units] = {2:.2f} [New Units]\n" \
                          "(at lambda = {0:0.4e})".format(*message_param)
            else:
                message = "Original Units: [{1}]\n\n" \
                          "New Units: [{3}]\n\n" \
                          "1.0 [Original Units] = {2:0.2e} [New Units]\n" \
                          "(at lambda = {0:0.4e})".format(*message_param)
        else:
            try:
                pixel_area = pixel_area.to("arcsec2 / pix")
                if isinstance(pixel_area, float):
                    pixel_area = pixel_area * u.arcsec ** 2 / u.pixel
            except (ValueError, u.UnitConversionError):
                pass
            message_param = (wave, self._original_unit.to_string(), new_value, unit_base, pixel_area)
            if 0.01 <= abs(new_value) <= 1000:
                message = "Original Units: [{1}]\n\n" \
                          "New Units: [{3}]\n\n" \
                          "Pixel Area (Scale): {4:.2e}\n\n" \
                          "1.0 [Original Units] = {2:.2f} [New Units]\n" \
                          "(at lambda = {0:0.4e})".format(*message_param)
            else:
                message = "Original Units: [{1}]\n\n" \
                          "New Units: [{3}]\n\n" \
                          "Pixel Area (Scale): {4:.2e}\n\n" \
                          "1.0 [Original Units] = {2:0.2e} [New Units]\n" \
                          "(at lambda = {0:0.4e})".format(*message_param)
        self.message_box.setText(message)

    def _on_flux_combo_change(self, index):
        """
        Callback for flux combo
        :param index:
        :return:
        """
        current_string = self.flux_combo.currentText()
        flux_unit_str = self.spectral_flux_density.to_string()
        if current_string != flux_unit_str:
            self.power_input.setText("0")
        else:
            power = self._get_current_power()
            self.power_input.setText(str(power))

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets. Set wave attribute to current
        wavelength.

        Layouts:
            10^[QLineEdit] X [QComboBox] / [QComboBox]
                        or
            10^[QLineEdit] X [QComboBox]

        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        power = self._get_current_power()

        unit_layout.addWidget(QLabel("10^"))
        power_input = QLineEdit(str(power))
        power_input.setFixedWidth(30)
        self.power_input = power_input
        power_input.textChanged.connect(self._update_message)
        unit_layout.addWidget(power_input)

        unit_layout.addWidget(QLabel("   X   "))

        flux_unit_str = self.spectral_flux_density.to_string()
        flux_options = FLUX_UNIT_REGISTRY.get_unit_list(current_unit=flux_unit_str)
        if flux_unit_str in flux_options:
            index = flux_options.index(flux_unit_str)
        else:
            index = find_unit_index(flux_options, flux_unit_str)
            if index is None:
                flux_options.append(flux_unit_str)
                index = flux_options.index(flux_unit_str)
        flux_combo = QComboBox()
        flux_combo.addItems(flux_options)
        flux_combo.setCurrentIndex(index)
        flux_combo.currentIndexChanged.connect(self._on_flux_combo_change)
        flux_combo.currentIndexChanged.connect(self._update_message)
        self.flux_combo = flux_combo
        unit_layout.addWidget(flux_combo)

        if self.area is not None:
            division = QLabel("   /   ")
            unit_layout.addWidget(division)

            area_str = self.area.to_string()
            if self.controller is None:
                no_pixel_area = True
            else:
                no_pixel_area = self.controller.pixel_area is None

            if self.area.decompose() == u.pix.decompose() and no_pixel_area:
                area_options = AREA_UNIT_REGISTRY.get_unit_list(pixel_only=True)
            elif 'solid angle' in self.area.physical_type and no_pixel_area:
                area_options = AREA_UNIT_REGISTRY.get_unit_list(solid_angle_only=True)
            else:
                area_options = AREA_UNIT_REGISTRY.get_unit_list()

            if area_str in area_options:
                index = area_options.index(area_str)
            else:
                index = find_unit_index(area_options, area_str)
                if index is None:
                    area_options.append(area_str)
                    index = area_options.index(area_str)
            area_combo = QComboBox()
            area_combo.width()
            area_combo.addItems(area_options)
            area_combo.setCurrentIndex(index)
            area_combo.currentIndexChanged.connect(self._update_message)
            self.area_combo = area_combo
            unit_layout.addWidget(area_combo)
        unit_layout.addStretch(1)
        unit_layout.setSpacing(0)

        if self.message_box is not None:
            self._update_message()

        if gui is not None:
            cubeviz_layout = gui.cubeviz_layout
            if cubeviz_layout is not None:
                self.wave = cubeviz_layout.get_wavelength()
        return unit_layout


class UnknownUnit(CubeVizUnit):
    """
    CubeVizUnit for any unit with a unit_string.
    Does not need to have an astropy unit.
    """
    def __init__(self, unit, unit_string):
        super(UnknownUnit, self).__init__(unit, unit_string)
        self._type = "UnknownUnit"

        if unit is not None:
            self.is_convertible = True

        self.options_combo = None

    def change_units(self):
        """
        This function is called when accept is pressed on the
        unit conversion gui
        :return: (bool) True if unit update is successful
        """
        if not self.is_convertible:
            return

        new_unit_string = self.options_combo.currentText()

        self._unit_string = new_unit_string
        self._unit = u.Unit(new_unit_string)

        return True

    def reset_widgets(self):

        self.message_box = None
        self.options_combo = None

    def _update_message(self):
        """
        Callback for when options in conversion gui change. Message
        is set to info about the units a preview of the conversion.
        """
        if self.message_box is None:
            return

        if not self.is_convertible:
            self.message_box.setText("")

        new_value = 1.0

        new_unit_string = self.options_combo.currentText()
        new_unit = u.Unit(new_unit_string)

        new_value *= self._original_unit.to(new_unit)

        if isinstance(new_value, u.Quantity):
            new_value = new_value.value

        message_param = (self._original_unit, new_unit_string, new_value)

        if 0.01 <= abs(new_value) <= 1000:
            message = "Original Units: [{0}]\n" \
                      "New Units: [{1}]\n" \
                      "1.0 [Original Units] = {2:.2f} [New Units]".format(*message_param)
        else:
            message = "Original Units: [{0}]\n" \
                      "New Units: [{1}]\n" \
                      "1.0 [Original Units] = {2:0.2e} [New Units]".format(*message_param)
        self.message_box.setText(message)

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets.

        Layouts:
            [QComboBox]
               or
            default_message

        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """

        if self.unit is None:
            default_message = "CubeViz can not convert this unit: {0}."
            default_message = default_message.format(self.unit_string)
            default_label = QLabel(default_message)
            unit_layout.addWidget(default_label)
            return unit_layout
        unit_str = self.unit.to_string()
        options = self.unit.find_equivalent_units()
        options = [i.to_string() for i in options]
        if unit_str not in options:
            options.append(unit_str)
        index = options.index(unit_str)
        combo = QComboBox()
        # combo.setFixedWidth(200)
        combo.addItems(options)
        combo.setCurrentIndex(index)
        combo.currentIndexChanged.connect(self._update_message)
        self.options_combo = combo
        unit_layout.addWidget(combo)
        self._update_message()
        return unit_layout


class NoneUnit(CubeVizUnit):
    """
    CubeVizUnit for components with no
    units and unit strings
    """
    def __init__(self):
        super(NoneUnit, self).__init__(None, "")
        self._type = "NoneUnit"

    def populate_unit_layout(self, unit_layout, gui=None):
        """
        Populate horizontal layout (living on conversion gui)
        with appropriate widgets.

        Layouts:
            default_message

        :param unit_layout: (QHBoxLayout) Horizontal layout
        :param gui: conversion gui
        :return: updated unit_layout
        """
        default_message = "No Units."
        default_label = QLabel(default_message)
        unit_layout.addWidget(default_label)
        return unit_layout


class ConvertFluxUnitGUI(QDialog):
    """
    GUI for unit conversions
    """
    def __init__(self, controller, parent=None):
        super(ConvertFluxUnitGUI, self).__init__(parent=parent)
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.title = "Unit Conversion"
        self.setMinimumSize(400, 270)

        self.cubeviz_layout = controller.cubeviz_layout
        self._hub = self.cubeviz_layout.session.hub

        self.controller = controller
        self.data = controller.data
        self.controller_components = controller._components

        self.current_unit = None

        self._init_ui()

    def _init_ui(self):
        # LINE 1: Data component drop down
        self.component_prompt = QLabel("Data Component:")
        self.component_prompt.setWordWrap(True)
        # Add the data component labels to the drop down, with the ComponentID
        # set as the userData:
        if self.parent is not None and hasattr(self.parent, 'data_components'):
            self.label_data = [(str(cid), cid) for cid in self.parent.data_components]
        else:
            self.label_data = [(str(cid), cid) for cid in self.data.visible_components]

        default_index = 0
        self.component_combo = QComboBox()
        self.component_combo.setFixedWidth(200)
        update_combobox(self.component_combo, self.label_data, default_index=default_index)
        self.component_combo.currentIndexChanged.connect(self.update_unit_layout)

        # hbl is short for Horizontal Box Layout
        hbl1 = QHBoxLayout()
        hbl1.addWidget(self.component_prompt)
        hbl1.addWidget(self.component_combo)
        hbl1.addStretch(1)

        # LINE 2: Unit conversion layout
        # This layout is filled by CubeVizUnit
        self.unit_layout = QHBoxLayout()  # this is hbl2

        # LINE 3: Message box
        self.message_box = QLabel("")
        hbl3 = QHBoxLayout()
        hbl3.addWidget(self.message_box)
        hbl3.addStretch(1)

        # Line 4: Buttons
        self.okButton = QPushButton("Convert Units")
        self.okButton.clicked.connect(self.call_main)
        self.okButton.setDefault(True)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.cancel)

        hbl4 = QHBoxLayout()
        hbl4.addStretch(1)
        hbl4.addWidget(self.cancelButton)
        hbl4.addWidget(self.okButton)

        vbl = QVBoxLayout()
        vbl.addLayout(hbl1)
        vbl.addLayout(self.unit_layout)
        vbl.addLayout(hbl3)
        vbl.addLayout(hbl4)
        self.setLayout(vbl)
        self.vbl = vbl

        self.update_unit_layout(default_index)

        self.show()

    def update_unit_layout(self, index):
        """
        Call back for component selection drop down.
        """
        component_id = str(self.component_combo.currentData())

        # STEP1: Clean up widgets from last component
        widgets = (self.unit_layout.itemAt(i) for i in range(self.unit_layout.count()))
        for w in widgets:
            if isinstance(w, QSpacerItem):
                self.unit_layout.removeItem(w)
                continue
            elif isinstance(w, QWidgetItem):
                w = w.widget()

            if hasattr(w, "deleteLater"):
                w.deleteLater()

        self.message_box.setText("")

        if self.current_unit:
            self.current_unit.reset_widgets()

        # STEP2: Add now component and connect to CubeVizUnit
        #        so that new widgets are populated.
        if component_id in self.controller_components:
            cubeviz_unit = self.controller_components[component_id]
            self.current_unit = cubeviz_unit
            cubeviz_unit.set_message_box(self.message_box)
            cubeviz_unit.populate_unit_layout(self.unit_layout, self)
            if cubeviz_unit.is_convertible:
                self.okButton.setEnabled(True)
            else:
                self.okButton.setEnabled(False)
        else:
            self.current_unit = None
            default_message = "CubeViz can not convert this unit."
            default_label = QLabel(default_message)
            self.unit_layout.addWidget(default_label)
            self.okButton.setEnabled(False)

        self.unit_layout.update()
        self.vbl.update()

    def call_main(self):
        """
        Calls CubeVizUnit.change_units to finalize
        conversions. Updates plots with new units.
        :return:
        """
        success = self.current_unit.change_units()
        if not success:
            # Todo: Warning should pop up
            return

        component_id = self.component_combo.currentData()
        self.data.get_component(component_id).units = self.current_unit.unit_string
        msg = FluxUnitsUpdateMessage(self, self.current_unit.unit, component_id)
        self._hub.broadcast(msg)
        self.close()

    def cancel(self):
        self.close()


class FluxUnitController:
    """
    Main controller for flux units.
    One FluxUnitController can only
    handle one glue data.
    """
    def __init__(self, cubeviz_layout=None):
        self.cubeviz_layout = cubeviz_layout

        # Load up configurations from yaml file
        with open(DEFAULT_FLUX_UNITS_CONFIGS, 'r') as yamlfile:
            cfg = yaml.load(yamlfile)

        # Define and add new units to astropy
        for new_unit_key in cfg["new_units"]:
            new_unit = cfg["new_units"][new_unit_key]
            try:
                u.Unit(new_unit["name"])
            except ValueError:
                if "base" in new_unit:
                    new_astropy_unit = u.def_unit(new_unit["name"], u.Unit(new_unit["base"]))
                else:
                    new_astropy_unit = u.def_unit(new_unit["name"])
                self.register_new_unit(new_astropy_unit)

        self._define_new_physical_types()

        # Formatted units
        self.formatted_units = cfg["formatted_units"]

        self.data = None
        self.wcs = None
        self._components = {}

    @property
    def components(self):
        return self._components

    @property
    def pixel_area(self):
        if self.wcs is None:
            return None
        try:
            top_unit = u.Unit(self.wcs.wcs.cunit[0])
            pixel_area = proj_plane_pixel_area(self.wcs) * (top_unit ** 2) / u.pix
            u.spectral_density.pixel_area = pixel_area
            return pixel_area
        except (ValueError, AttributeError):
            return None

    @staticmethod
    def register_new_unit(new_unit):
        """
        Add new unit to astropy.units
        :param new_unit: astropy unit
        """
        u.add_enabled_units(new_unit)
        if FLUX_UNIT_REGISTRY.is_compatible(new_unit):
            FLUX_UNIT_REGISTRY.add_unit(new_unit)
        if new_unit.decompose() == u.pix.decompose() or \
                'solid angle' in new_unit.physical_type:
            AREA_UNIT_REGISTRY.add_unit(new_unit)

    @staticmethod
    def _define_new_physical_types():
        """
        Two new types of units defined:
            - SFD_over_solid_angle = spectral_flux_density / (degree^2)
            - SFD_over_pix = spectral_flux_density / pixel
        """
        new_physical_types = [
            [(u.Jy / u.degree ** 2), 'SFD_over_solid_angle'],
            [(u.Jy / u.pix), 'SFD_over_pix']
        ]
        for model_unit, name in new_physical_types:
            try:
                u.def_physical_type(model_unit, name)
            except ValueError:
                continue

    @staticmethod
    def _sfd_over_solid_angle_to_cubeviz(unit, unit_string):
        """
        Decompose sfd_over_solid_angle to CubeVizUnit
        NB: SFD_over_solid_angle = spectral_flux_density / (degree^2)
        :param unit: astropy unit
        :param unit_string: astropy unit
        :return: CubeVizUnit
        """
        power = _get_power(unit.scale)
        unit_list = unit.bases
        power_list = unit.powers

        index = None
        sfd_unit = None
        for i, un in enumerate(unit_list):
            if 'solid angle' == un.physical_type or \
                    'angle' == un.physical_type:
                index = i
            else:
                if sfd_unit is None:
                    sfd_unit = un ** power_list[i]
                else:
                    sfd_unit *= un ** power_list[i]

        if index is not None:
            angle_unit = unit_list[index] ** abs(power_list[index])
            if power == 0:
                sfd_unit = u.Unit(unit.scale * sfd_unit)

            cubeviz_unit = SpectralFluxDensity(unit, unit_string,
                                               power, sfd_unit, angle_unit)
        else:
            cubeviz_unit = UnknownUnit(unit, unit_string)
        return cubeviz_unit

    @staticmethod
    def _sfd_over_pix_to_cubeviz(unit, unit_string):
        """
        Decompose sfd_over_pix to CubeVizUnit
        NB: SFD_over_pix = spectral_flux_density / pixel
        :param unit: astropy unit
        :param unit_string: astropy unit
        :return: CubeVizUnit
        """
        power = _get_power(unit.scale)
        unit_list = unit.bases
        power_list = unit.powers

        index = None
        sfd_unit = None
        for i, un in enumerate(unit_list):
            if un.decompose() == u.pix.decompose():
                index = i
            else:
                if sfd_unit is None:
                    sfd_unit = un ** power_list[i]
                else:
                    sfd_unit *= un ** power_list[i]

        if index is not None:
            pix_unit = unit_list[index] ** abs(power_list[index])
            if power == 0:
                sfd_unit = u.Unit(unit.scale * sfd_unit)
            cubeviz_unit = SpectralFluxDensity(unit, unit_string,
                                               power, sfd_unit, pix_unit)
        else:
            cubeviz_unit = UnknownUnit(unit, unit_string)
        return cubeviz_unit

    @staticmethod
    def string_to_unit(unit_string):
        """
        Unit str -> astropy unit
        :param unit_string: str
        :return: astropy unit. None otherwise
        """
        try:
            if unit_string:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', AstropyWarning)
                    astropy_unit = u.Unit(unit_string)
                return astropy_unit
        except ValueError:
            print("Warning: Failed to convert {0} to an astropy unit.".format(unit_string))
        return None

    @staticmethod
    def unit_to_string(unit):
        """
        astropy unit -> unit str
        :param unit: astropy unit
        :return: str
        """
        if unit is None:
            return ""
        elif isinstance(unit, str):
            return unit
        elif isinstance(unit, u.UnitBase):
            return unit.to_string()
        elif isinstance(unit, u.quantity.Quantity):
            return unit.unit.to_string()
        else:
            raise ValueError("Invalid input unit type {0}.".format(type(unit)))

    def add_component_unit(self, component_id, unit=None):
        """
        Add or update componet unit.
        :param component_id: component id or str
        :param unit: string or astropy unit
        :return: CubeVizUnit
        """
        component_id = str(component_id)
        unit_string = self.unit_to_string(unit)

        # IF: empty unit or no unit
        if not unit_string:
            cubeviz_unit = NoneUnit()
        # ELSE IF: formatted unit
        elif unit_string in self.formatted_units:
            registered_unit = self.formatted_units[unit_string]
            if "astropy_unit_string" in registered_unit:
                unit_string = registered_unit["astropy_unit_string"]
            astropy_unit = self.string_to_unit(unit_string)
            spectral_flux_density = self.string_to_unit(registered_unit["spectral_flux_density"])
            if "area" in registered_unit:
                area = self.string_to_unit(registered_unit["area"])
            else:
                area = None

            numeric = spectral_flux_density.scale
            power = _get_power(numeric)
            if power != 0:
                spectral_flux_density /= u.Unit(numeric)
            cubeviz_unit = SpectralFluxDensity(astropy_unit, unit_string,
                                               power, spectral_flux_density, area,
                                               is_formatted=True)
        else:
            astropy_unit = self.string_to_unit(unit_string)
            # IF: no astropy unit component
            if astropy_unit is None:
                cubeviz_unit = UnknownUnit(astropy_unit, unit_string)
            # ELSE IF: astropy unit type is spectral_flux_density / solid_angle
            elif 'SFD_over_solid_angle' in astropy_unit.physical_type:
                cubeviz_unit = self._sfd_over_solid_angle_to_cubeviz(astropy_unit, unit_string)
            # ELSE IF: astropy unit type is spectral_flux_density / pixel
            elif 'SFD_over_pix' in astropy_unit.physical_type:
                cubeviz_unit = self._sfd_over_pix_to_cubeviz(astropy_unit, unit_string)
            # ELSE IF: astropy unit type is spectral_flux_density
            elif 'spectral flux density' in astropy_unit.physical_type:
                spectral_flux_density = astropy_unit
                numeric = spectral_flux_density.scale
                power = _get_power(numeric)
                if power != 0:
                    spectral_flux_density /= u.Unit(numeric)
                cubeviz_unit = SpectralFluxDensity(astropy_unit, unit_string,
                                                   power, spectral_flux_density,
                                                   area=None)
            # ELSE: astropy unit type is not special
            else:
                cubeviz_unit = UnknownUnit(astropy_unit, unit_string)

        cubeviz_unit.controller = self
        self._components[component_id] = cubeviz_unit
        return cubeviz_unit

    def remove_component_unit(self, component_id):
        """
        Remove component from controller
        :param component_id: component id or str
        """
        component_id = str(component_id)
        if component_id in self._components:
            del self._components[component_id]

    def get_component_unit(self, component_id, cubeviz_unit=False):
        """
        Get component units
        :param component_id: component id or str
        :param cubeviz_unit: if True: return CubeVizUnit
                             else: return astropy unit
        :return: CubeVizUnit, astropy unit or None
        """
        component_id = str(component_id)
        if component_id in self._components:
            if cubeviz_unit:
                return self._components[component_id]
            else:
                return self._components[component_id].unit
        return None

    def set_data(self, data):
        """
        Add glue data and its components to controller.
        glue components' units attribute should contain a
        string with desired units for that component. This
        controller will use the string in that attribute
        to construct and assign units.
        :param data: glue data
        :return:
        """
        self.data = data
        self._components = {}
        for comp in data.visible_components:
            self.add_component_unit(comp, data.get_component(comp).units)

        wcs = data.coords.wcs
        if wcs is not None:
            self.wcs = wcs
        self.pixel_area

    def converter(self, parent=None):
        """
        Launch Converter GUI
        :param parent: application
        :return: ConvertFluxUnitGUI instance
        """
        ex = ConvertFluxUnitGUI(self, parent)
        return ex



from __future__ import absolute_import, division, print_function

from qtpy.QtCore import Qt
from qtpy import QtGui
from qtpy.QtWidgets import (
    QDialog, QApplication, QPushButton,
    QLabel, QWidget, QHBoxLayout, QVBoxLayout, QLineEdit, QComboBox
)

from astropy.stats import sigma_clip
from glue.utils.qt import load_ui

import numpy as np
import re
import os

from ..controls.wavelengths import REST_WAVELENGTH_TEXT, OBS_WAVELENGTH_TEXT

class WavelengthUI(QDialog):
    def __init__(self, wavelength_controller, parent=None):
        super(WavelengthUI,self).__init__(parent)

        self.wavelength_controller = wavelength_controller

        # Load the widget from the UI file.
        self.ui = load_ui('wavelength.ui', self,
                          directory=os.path.dirname(__file__))

        # Add the wavelength units to the combo box
        self.ui.wavelengthunits_combobox.addItems(
                self.wavelength_controller.unit_titles)

        current_units = self.wavelength_controller.current_units
        self.ui.wavelengthunits_combobox.setCurrentIndex(
            self.wavelength_controller.units.index(current_units))

        # Add the wavelength display to the combo box
        self.ui.wavelengthdisplay_combobox.addItems(
                [OBS_WAVELENGTH_TEXT, REST_WAVELENGTH_TEXT])

        # Set default values based on what was previously selected
        # This should be before the callbacks just so we don't fire anything
        z = wavelength_controller.redshift_z
        self.ui.wavelengthdisplay_combobox.setCurrentIndex(0 if z == 0 else 1)
        self.ui.redshift_text.setText(str(z))

        # We want to initially disable the redshift part if there
        # is no redshift, then when they switch to "Rest Wavelength"
        # these will be enabled.
        self.ui.redshift_label.setDisabled(z == 0)
        self.ui.redshift_text.setDisabled(z == 0)

        # Setup the callbacks on the UI
        self.ui.wavelengthdisplay_combobox.currentIndexChanged.connect(
                self._wavelengthdisplay_selection_change)
        self.ui.calculate_button.clicked.connect(self._calculate_callback)
        self.ui.cancel_button.clicked.connect(self._cancel_callback)

        self.ui.show()

    def do_calculation(self, wavelength_redshift, wavelength_units):
        """
        Actual calculate function so that we can use this in the
        testing scripts as well.

        :param wavelength_redshift:  z-value
        :param wavelength_units: text representation of the units
        :return: nuttin
        """
        self.wavelength_controller.update_units(wavelength_units)
        self.wavelength_controller.update_redshift(wavelength_redshift)

    def _calculate_callback(self):

        # Reset the errors
        self.ui.error_text.setText(' ')
        self.ui.error_text.setStyleSheet("color: rgba(0, 0, 0, 128)")
        self.ui.redshift_label.setStyleSheet("color: rgba(0, 0, 0, 128)")

        # Check the redshift value if we are using the Obs wavelengths
        if REST_WAVELENGTH_TEXT == self.wavelengthdisplay_combobox.currentText():
            redshift = self.ui.redshift_text.text().strip()

            try:
                redshift = float(redshift)
            except Exception as e:
                self.ui.redshift_label.setStyleSheet("color: rgba(255, 0, 0, 128)")
                self.ui.error_text.setText('Redshift value {} does not appear to be a number'.format(redshift))
                return
        else:
            redshift = 0.0

        # Get the units based on the index in the units combo box.
        index = self.ui.wavelengthunits_combobox.currentIndex()
        units = self.wavelength_controller.units[index]

        # Do the actual calculation
        self.do_calculation(wavelength_redshift=redshift, wavelength_units=units)

        self.close()

    def _cancel_callback(self, caller=0):
        """
        Cancel callback when the person hits the cancel button

        :param caller:
        :return:
        """
        self.close()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.cancel_callback()

    def _wavelengthdisplay_selection_change(self, index):
        """
        Callback for a change on the region selection combo box.

        :param newvalue:
        :return:
        """
        newvalue = self.ui.wavelengthdisplay_combobox.currentText()

        # Hide the redshift stuff if Observed wavelength is selected
        self.ui.redshift_label.setDisabled(newvalue == OBS_WAVELENGTH_TEXT)
        self.ui.redshift_text.setDisabled(newvalue == OBS_WAVELENGTH_TEXT)



from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import ndimage

from astropy import convolution

from glue.core import Data, Subset, Component
from glue.core.coordinates import coordinates_from_header, WCSCoordinates
from glue.core.exceptions import IncompatibleAttribute
from glue.utils.qt import update_combobox

from spectral_cube import SpectralCube, BooleanArrayMask

from qtpy.QtCore import Qt, Signal, QThread
from qtpy.QtWidgets import (
    QDialog, QApplication, QPushButton, QProgressBar,
    QLabel, QWidget, QDockWidget, QHBoxLayout, QVBoxLayout,
    QComboBox, QMessageBox, QLineEdit, QRadioButton
)


class AbortException(Exception):
    """
    Custom exception to indicate a calculation abort.
    """
    pass


class WorkerThread(QThread):
    """
    Custom QThread for SmoothCube
    """

    success_signal = Signal()  # Smoothing is done
    error_signal = Signal(Exception)  # Smoothing failed

    def __init__(self, smooth_cube, parent=None):
        super(WorkerThread, self).__init__(parent)
        self.smooth_cube = smooth_cube

        self.success_signal.connect(self.smooth_cube.thread_callback)
        self.error_signal.connect(self.smooth_cube.thread_error_handler)

    def run(self):
        try:
            success_flag = self.smooth_cube.thread_function()
            if success_flag:
                self.success_signal.emit()
        except Exception as e:
            if not isinstance(e, AbortException):
                self.error_signal.emit(e)


class SmoothCube(object):
    """
    SmoothCube is a wrapper for SpectralCube. It is designed to
    operate in CubeViz and glue.
    It saves a registry of available kernels and executes
    smoothing operations. It has the ability to use QThreads
    when working in gui mode.
    """

    def __init__(self, data=None, smoothing_axis=None, kernel_type=None, kernel_size=None,
                 component_id=None, output_label=None, output_as_component=False, parent=None):
        self.data = data  # Glue data to be smoothed
        self.smoothing_axis = smoothing_axis  # spectral vs spatial
        self.kernel_type = kernel_type  # Type of kernel, a key in kernel_registry
        self.kernel_size = kernel_size  # Size of kernel in pix
        self.component_id = component_id  # Glue data component to smooth over
        self.component_unit = ''  # Units of component
        self.output_label = output_label  # Output label
        self.output_as_component = output_as_component  # Add output component to self.data
        self.kernel_registry = self.load_kernel_registry()  # A list of kernels and params

        self.parent = parent  # If gui is going to be used, parent of new gui

        # Vars for multi-threading smoothing
        self.is_multi_threading = False  # Using multi-threading?
        self.is_active = False  # Is thread active
        self.abort_window = None  # Abort window gui
        self.thread = None # QThread
        self.thread_cube = None  # Input SpectralCube
        self.thread_result = None  # Temporary storage for thread output

    @staticmethod
    def load_kernel_registry():
        """
        Nested dict to store available kernel options.

        kernel_registry: Main registry
            kernel_type: Type of kernel
                name: Display name
                unit_label: Display units of kernel size
                size_prompt: User prompt for size of kernel
                size_dimension: Dimension of kernel (width, radius, etc..)
                axis: Available axes. Each must have a kernel.
                spatial: 2D astropy.convolution.kernel else None.
                spectral: 1D astropy.convolution.kernel else None.

        :return: kernel_registry
        """
        kernel_registry = {
            "box": {"name": "Box",
                    "unit_label": "Spaxel",
                    "size_prompt": "Width of Kernel:",
                    "size_dimension": "Width",
                    "axis": ["spatial", "spectral"],
                    "spatial": convolution.Box2DKernel,
                    "spectral": convolution.Box1DKernel},
            "gaussian": {"name": "Gaussian",
                         "unit_label": "Spaxel",
                         "size_prompt": "Standard Deviation of Kernel:",
                         "size_dimension": "Sigma",
                         "axis": ["spatial", "spectral"],
                         "spatial": convolution.Gaussian2DKernel,
                         "spectral": convolution.Gaussian1DKernel},
            "trapezoid": {"name": "Trapezoid",
                          "unit_label": "Spaxel",
                          "size_prompt": "Width of Kernel:",
                          "size_dimension": "Width",
                          "axis": ["spectral"],
                          "spectral": convolution.Trapezoid1DKernel},
            "trapezoiddisk": {"name": "Trapezoid Disk",
                              "unit_label": "Spaxel",
                              "size_prompt": "Radius of Kernel:",
                              "size_dimension": "Radius",
                              "axis": ["spatial"],
                              "spatial": convolution.TrapezoidDisk2DKernel},
            "airydisk": {"name": "Airy Disk",
                         "unit_label": "Spaxel",
                         "size_prompt": "Radius of Kernel:",
                         "size_dimension": "Radius",
                         "axis": ["spatial"],
                         "spatial": convolution.AiryDisk2DKernel},
            "tophat": {"name": "Top Hat",
                       "unit_label": "Spaxel",
                       "size_prompt": "Radius of Kernel:",
                       "size_dimension": "Radius",
                       "axis": ["spatial"],
                       "spatial": convolution.Tophat2DKernel},
            "median": {"name": "Median",
                       "unit_label": "Spaxel",
                       "size_prompt": "Width of Kernel:",
                       "size_dimension": "Width",
                       "axis": ["spatial", "spectral"],
                       "spatial": None,
                       "spectral": None}
        }
        return kernel_registry

    def get_kernel_registry(self):
        return self.kernel_registry

    def get_kernel(self):
        """
        Gets an kernel using the saved parameters
        :return: astropy.convolution.kernel
        """
        return self.kernel_registry[self.kernel_type][self.smoothing_axis](self.kernel_size)

    def get_kernel_size_prompt(self, kernel_type=None):
        """kernel_type -> size_prompt"""
        if kernel_type is None:
            if self.kernel_type is None:
                return None
            kernel_type = self.kernel_type
        if kernel_type not in self.kernel_registry:
            return None
        return self.kernel_registry[kernel_type]["size_prompt"]

    def get_kernel_unit(self, kernel_type=None):
        """kernel_type -> unit_label"""
        if kernel_type is None:
            if self.kernel_type is None:
                return None
            kernel_type = self.kernel_type
        if kernel_type not in self.kernel_registry:
            return None
        return self.kernel_registry[kernel_type]["unit_label"]

    def get_kernel_size_dimension(self, kernel_type=None):
        """kernel_type -> size_dimension"""
        if kernel_type is None:
            if self.kernel_type is None:
                return None
            kernel_type = self.kernel_type
        if kernel_type not in self.kernel_registry:
            return None
        return self.kernel_registry[kernel_type]["size_dimension"]

    def name_to_kernel_type(self, name):
        """kernel name -> kernel_type"""
        for k in self.kernel_registry:
            if name == self.kernel_registry[k]["name"]:
                return k
        return None

    def kernel_type_to_name(self, kernel_type):
        """kernel_type -> kernel name"""
        for k in self.kernel_registry:
            if kernel_type == k:
                return self.kernel_registry[k]["name"]
        return None

    def get_glue_wcs(self):
        """
        Get WCS of glue Data.
        :return: astropy.wcs.WCS
        :raises: Exception WCS not found
        """
        coords = self.data.coords
        if isinstance(coords, WCSCoordinates):
            return coords.wcs
        raise Exception("WCS information was not provided.")

    def get_glue_mask(self):
        """Function make/extract mask of glue Data"""
        if isinstance(self.data, Subset):
            try:
                mask = self.data.to_mask()
                return mask
            except IncompatibleAttribute:
                pass
        d = self.data[self.component_id]
        mask = np.empty_like(d)
        mask.fill(True)
        mask = mask.astype(bool)
        return mask

    def data_to_cube(self):
        """Glue Data -> SpectralCube"""
        if self.component_id is None:
            raise Exception("component_id was not provided.")
        wcs = self.get_glue_wcs()
        data_array = self.data[self.component_id]
        mask = BooleanArrayMask(
            mask=self.get_glue_mask(),
            wcs=wcs)
        self.component_unit = self.data.get_component(self.component_id).units
        return SpectralCube(data=data_array, wcs=wcs, mask=mask)

    def cube_to_data(self, cube,
                     output_label=None,
                     output_component_id=None):
        """
        Convert SpectralCube to final output.
        self.output_as_component is checked here.
        if self.output_as_component:
            add new component to self.data
        else:
            create new data and return it.
        :param cube: SpectralCube
        :param output_label: Name of new Data.
        :param output_component_id: label of new component
        :return:
        """
        original_data = self.data
        new_component = Component(cube._data.copy(), self.component_unit)
        if self.output_as_component:
            original_data.add_component(new_component, output_component_id)
            return None
        else:
            new_data = Data(label=output_label)
            new_data.coords = coordinates_from_header(cube.header)
            new_data.add_component(new_component, output_component_id)
            return new_data

    def unique_output_component_id(self):
        unit_label = self.kernel_registry[self.kernel_type]["unit_label"].lower()
        if self.kernel_size == 1:
            kernel_size_text = "{0}_{1}".format(self.kernel_size, unit_label)
        else:
            kernel_size_text = "{0}_{1}s".format(self.kernel_size, unit_label)

        name_tail = "_Smoothed(" + ", ".join([
                                            self.kernel_type_to_name(self.kernel_type),
                                            self.smoothing_axis.title(),
                                            kernel_size_text]) + ")"

        if self.output_label is None:
            output_component_id = self.component_id + name_tail
        else:
            output_component_id = self.output_label

        taken = [str(i) for i in self.data.component_ids()]
        index_list = []
        for name in taken:
            if output_component_id in name:
                if output_component_id == name:
                    index_list.append(0)
                else:
                    try:
                        idx = int(name.split("_")[-1])
                        index_list.append(idx)
                    except ValueError:
                        pass
        if len(index_list) > 0:
            idx = max(index_list) + 1
            return output_component_id + "_" + str(idx)
        return output_component_id

    def output_data_name(self):
        unit_label = self.kernel_registry[self.kernel_type]["unit_label"].lower()
        if self.kernel_size == 1:
            kernel_size_text = "{0}_{1}".format(self.kernel_size, unit_label)
        else:
            kernel_size_text = "{0}_{1}s".format(self.kernel_size, unit_label)

        name_tail = "_Smoothed(" + ", ".join([
                                            self.kernel_type_to_name(self.kernel_type),
                                            self.smoothing_axis.title(),
                                            kernel_size_text]) + ")"

        if self.output_label is None:
            return self.data.label + name_tail
        else:
            return self.output_label

    def smooth_cube(self, preview=False):
        """
        Main (serial) smoothing function that follows the following steps:
        1) Convert data to SpectralCube
        2) Obtain Kernel using saved parameters (if applicable)
        3) Smooth cube
        4) Generate name based on saved parameters
        5) Output component or Data
        :return: glue.core.Data or None
        """
        cube = self.data_to_cube()

        if "median" == self.kernel_type:
            if "spatial" == self.smoothing_axis:
                new_cube = cube.spatial_smooth_median(self.kernel_size)
            else:
                new_cube = cube.spectral_smooth_median(self.kernel_size)
        else:
            kernel = self.get_kernel()
            if "spatial" == self.smoothing_axis:
                new_cube = cube.spatial_smooth(kernel)
            else:
                new_cube = cube.spectral_smooth(kernel)

        if self.output_as_component:
            output_component_id = self.unique_output_component_id()
            output = self.cube_to_data(new_cube, output_component_id=output_component_id)
        elif preview:
            return cube._data.copy()
        else:
            output_label = self.output_data_name()
            output = self.cube_to_data(new_cube,
                                       output_label=output_label,
                                       output_component_id=self.component_id)

        return output

    def multi_threading_smooth(self):
        """
        Prepares data and starts worker thread.
        Overall steps accomplished:
            1) Convert data to SpectralCube and start thread
        """
        cube = self.data_to_cube()

        # Handshake b/w SpectralCube and AbortWindow
        if "spectral" == self.smoothing_axis:
            self.abort_window.init_pb(0, cube.shape[1]*cube.shape[2])
        else:
            self.abort_window.init_pb(0, cube.shape[0])

        self.thread_cube = cube
        self.thread = WorkerThread(self, self.parent)
        self.thread.start()

    def thread_function(self):
        """
        On-thread function that executes smoothing
        Overall steps accomplished:
            2) Obtain Kernel using saved parameters (if applicable)
            3) Smooth cube
        :return: bool: True if output is added to self.thread_result
        :raises: Exception: if output is not an instance of SpectralCube
        """
        success = True

        cube = self.thread_cube
        update_function = self.abort_window.update_pb
        if "median" == self.kernel_type:
            if "spatial" == self.smoothing_axis:
                new_cube = cube.spatial_smooth_median(self.kernel_size, update_function=update_function)
            else:
                new_cube = cube.spectral_smooth_median(self.kernel_size, update_function=update_function)
        else:
            kernel = self.get_kernel()
            if "spatial" == self.smoothing_axis:
                new_cube = cube.spatial_smooth(kernel, update_function=update_function)
            else:
                new_cube = cube.spectral_smooth(kernel, update_function=update_function)

        if isinstance(new_cube, SpectralCube):
            self.thread_result = new_cube
            return success
        else:
            raise Exception("Unexpected return type from SpectralCube.")

    def thread_callback(self):
        """
        Callback function for worker thread.
        Overall steps accomplished:
            4) Generate name based on saved parameters
            5) Output as component ONLY
        """
        output_component_id = self.unique_output_component_id()
        output = self.cube_to_data(self.thread_result, output_component_id=output_component_id)
        self.abort_window.smoothing_done(output_component_id)

    def thread_error_handler(self, exception):
        self.abort_window.print_error(exception)
        raise exception

    def gui(self):
        """Call smoothing gui and add output as component"""
        ex = SelectSmoothing(self.data, self.parent)

    def preview_smoothing(self, data):
        if "median" == self.kernel_type:
            return ndimage.filters.median_filter(data, self.kernel_size)
        else:
            kernel = self.get_kernel()
            return convolution.convolve(data, kernel, normalize_kernel=True)

    def get_preview_title(self):
        title = "Smoothing Preview: "
        title += self.kernel_type_to_name(self.kernel_type)
        size_dimension = self.get_kernel_size_dimension(self.kernel_type)
        unit_label = self.kernel_registry[self.kernel_type]["unit_label"].lower()
        if self.kernel_size == 1:
            title += "({0} = {1} {2})".format(size_dimension, self.kernel_size, unit_label)
        else:
            title += "({0} = {1} {2}s)".format(size_dimension, self.kernel_size, unit_label)
        return title


class AbortWindow(QDialog):
    """
    Displays busy message and provides abort button.
    The class serves SmoothCube, WorkerThread and SelectSmoothing.
    """

    def __init__(self, parent=None):
        """
        init abort or notification ui.
        Displays while smoothing freezes the application.
        Allows abort button to be added if needed.
        """
        super(AbortWindow, self).__init__(parent)
        self.setModal(False)
        self.setWindowFlags(self.windowFlags() | Qt.Tool | Qt.WindowStaysOnTopHint)

        self.parent = parent

        self.label_a_1 = QLabel("Executing smoothing algorithm.")
        self.label_a_2 = QLabel("This may take several minutes.")

        self.abort_button = QPushButton("Abort")
        self.abort_button.clicked.connect(self.abort)

        self.pb = QProgressBar(self)
        self.pb_counter = 0

        self.abort_flag = False

        # vbl is short for Vertical Box Layout
        vbl = QVBoxLayout()
        vbl.addWidget(self.label_a_1)
        vbl.addWidget(self.label_a_2)
        vbl.addWidget(self.pb)
        vbl.addWidget(self.abort_button)

        self.setLayout(vbl)

        self.show()

    def init_pb(self, start, end):
        """
        Init the progress bar
        :param start: Start Value
        :param end: End Value
        """
        self.pb.setRange(start, end)
        self.pb_counter = start

    def update_pb(self):
        """
        This function is called in the worker thread to
        update the progress bar and checks if the
        local class variable abort_flag is active.

        If the abort button is clicked, the main thread
        will set abort_flag to True. The next time the
        worker thread calls this function, a custom
        exception is raised terminating the calculation.
        The exception is handled by the WorkerThread class.
        :raises: AbortException: terminating smoothing calculation
        """
        if self.abort_flag:
            raise AbortException("Abort Calculation")
        self.pb_counter += 1
        self.pb.setValue(self.pb_counter)
        QApplication.processEvents()

    def abort(self):
        """Abort calculation"""
        self.abort_flag = True
        self.parent.clean_up()

    def smoothing_done(self, component_id=None):
        """Notify user success"""
        self.hide()
        if component_id is None:
            message = "The result has been added as a" \
                      " new component of the input Data." \
                      " The new component can be accessed" \
                      " in the viewer drop-downs."
        else:
            message = "The result has been added as" \
                      " \"{0}\" and can be selected" \
                      " in the viewer drop-down menu.".format(component_id)

        info = QMessageBox.information(self, "Success", message)
        self.clean_up()

    def print_error(self, exception):
        """Print error message"""

        if "signal only works in main thread" in str(exception):
            message = "Smoothing Failed!\n\n" + "Please update your SpectralCube package"
        else:
            message = "Smoothing Failed!\n\n" + str(exception)

        info = QMessageBox.critical(self, "Error", message)
        self.clean_up()

    def clean_up(self):
        self.parent.clean_up()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.abort()


class SelectSmoothing(QDialog):
    """
    SelectSmoothing launches a GUI and executes smoothing.
    Any output is added to the input data as a new component.
    """

    def __init__(self, data, parent=None, smooth_cube=None,
                 allow_preview=False, allow_spectral_axes=False):
        super(SelectSmoothing, self).__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.parent = parent
        self.title = "Smoothing Selection"

        self.data = data  # Glue data to be smoothed

        # Check if smooth object is the caller
        if smooth_cube is None:
            self.smooth_cube = SmoothCube(data=self.data)
        else:
            self.smooth_cube = smooth_cube

        self.allow_spectral_axes = allow_spectral_axes

        self.allow_preview = allow_preview
        self.is_preview_active = False  # Flag to show if smoothing preview is active

        self.abort_window = None  # Small window pop up when smoothing.

        self.component_id = None  # Glue data component to smooth over
        self.current_axis = None  # Selected smoothing_axis
        self.current_kernel_type = None  # Selected kernel type, a key in SmoothCube.kernel_registry
        self.current_kernel_name = None  # Name of selected kernel

        self._init_selection_ui()  # Format and show gui

    def _init_selection_ui(self):
        # LINE 1: Radio box spatial vs spectral axis
        self.axes_prompt = QLabel("Smoothing Axis:")
        self.axes_prompt.setMinimumWidth(150)

        self.spatial_radio = QRadioButton("Spatial")
        self.spatial_radio.setChecked(True)
        self.current_axis = "spatial"
        self.spatial_radio.toggled.connect(self.spatial_radio_checked)

        self.spectral_radio = QRadioButton("Spectral")
        self.spectral_radio.toggled.connect(self.spectral_radio_checked)

        # hbl is short for Horizontal Box Layout
        hbl1 = QHBoxLayout()
        hbl1.addWidget(self.axes_prompt)
        hbl1.addWidget(self.spatial_radio)
        hbl1.addWidget(self.spectral_radio)

        # LINE 2: Kernel Type prompt
        self.k_type_prompt = QLabel("Kernel Type:")
        self.k_type_prompt.setMinimumWidth(150)
        # Load kernel types + names and add to drop down
        self._load_options()
        self.combo = QComboBox()
        self.combo.setMinimumWidth(150)
        self.combo.addItems(self.options[self.current_axis])

        hbl2 = QHBoxLayout()
        hbl2.addWidget(self.k_type_prompt)
        hbl2.addWidget(self.combo)

        # LINE 3: Kernel size
        self.size_prompt = QLabel(self.smooth_cube.get_kernel_size_prompt(self.current_kernel_type))
        self.size_prompt.setWordWrap(True)
        self.size_prompt.setMinimumWidth(150)
        self.unit_label = QLabel(self.smooth_cube.get_kernel_unit(self.current_kernel_type))
        self.k_size = QLineEdit("1")  # Default Kernel size set here

        hbl3 = QHBoxLayout()
        hbl3.addWidget(self.size_prompt)
        hbl3.addWidget(self.k_size)
        hbl3.addWidget(self.unit_label)

        # LINE 4: Data component drop down
        self.component_prompt = QLabel("Data Component:")
        self.component_prompt.setWordWrap(True)
        self.component_prompt.setMinimumWidth(150)
        # Load component_ids and add to drop down

        # Add the data component labels to the drop down, with the ComponentID
        # set as the userData:

        if self.parent is not None and hasattr(self.parent, 'data_components'):
            labeldata = [(str(cid), cid) for cid in self.parent.data_components]
        else:
            labeldata = [(str(cid), cid) for cid in self.data.main_components()]

        self.component_combo = QComboBox()
        update_combobox(self.component_combo, labeldata)

        self.component_combo.setMaximumWidth(150)
        self.component_combo.setCurrentIndex(0)
        if self.allow_preview:
            self.component_combo.currentIndexChanged.connect(self.update_preview_button)

        hbl4 = QHBoxLayout()
        hbl4.addWidget(self.component_prompt)
        hbl4.addWidget(self.component_combo)

        # Line 5: Preview Message
        message = "Info: Smoothing previews are displayed on " \
                  "CubeViz's left and single image viewers."
        self.preview_message = QLabel(message)
        self.preview_message.setWordWrap(True)
        self.preview_message.hide()
        hbl5 = QHBoxLayout()
        hbl5.addWidget(self.preview_message)

        # LINE 6: preview ok cancel buttons
        self.previewButton = QPushButton("Preview Slice")
        self.previewButton.clicked.connect(self.call_preview)

        self.okButton = QPushButton("Smooth Cube")
        self.okButton.clicked.connect(self.call_main)
        self.okButton.setDefault(True)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.cancel)

        hbl6 = QHBoxLayout()
        hbl6.addStretch(1)
        if self.allow_preview:
            hbl6.addWidget(self.previewButton)
        hbl6.addWidget(self.cancelButton)
        hbl6.addWidget(self.okButton)

        # Add Lines to Vertical Layout
        # vbl is short for Vertical Box Layout
        vbl = QVBoxLayout()
        if self.allow_spectral_axes:
            vbl.addLayout(hbl1)
        vbl.addLayout(hbl2)
        vbl.addLayout(hbl3)
        vbl.addLayout(hbl4)
        vbl.addLayout(hbl5)
        vbl.addLayout(hbl6)

        self.setLayout(vbl)
        self.setMaximumWidth(330)

        # Connect kernel combo box to event handler
        self.combo.currentIndexChanged.connect(self.selection_changed)
        self.selection_changed(0)

        self.show()

    def _load_options(self):
        """Extract names + types of kernels from SmoothCube.kernel_registry"""
        kernel_registry = self.smooth_cube.get_kernel_registry()

        self.options = {"spatial": [], "spectral": []}
        for k in kernel_registry:
            axis = kernel_registry[k]["axis"]
            for a in axis:
                if "spatial" == a:
                    self.options["spatial"].append(kernel_registry[k]["name"])
                elif "spectral" == a:
                    self.options["spectral"].append(kernel_registry[k]["name"])
        self.options["spectral"].sort()
        self.options["spatial"].sort()
        self.current_kernel_name = self.options[self.current_axis][0]
        self.current_kernel_type = self.smooth_cube.name_to_kernel_type(self.options[self.current_axis][0])

    def selection_changed(self, i):
        """
        Update kernel type, units, etc... when
        kernel name changes in combo box.
        """
        keys = self.options[self.current_axis]
        name = keys[i]
        self.current_kernel_name = name
        self.current_kernel_type = self.smooth_cube.name_to_kernel_type(name)
        self.unit_label.setText(self.smooth_cube.get_kernel_unit(self.current_kernel_type))
        self.size_prompt.setText(self.smooth_cube.get_kernel_size_prompt(self.current_kernel_type))

    def spatial_radio_checked(self):
        self.current_axis = "spatial"
        self.update_preview_button()
        self.combo.clear()
        self.combo.addItems(self.options[self.current_axis])

    def spectral_radio_checked(self):
        self.current_axis = "spectral"
        self.update_preview_button()
        self.combo.clear()
        self.combo.addItems(self.options[self.current_axis])

    def input_validation(self):
        """
        Check if input will break Smoothing
        :return: bool: True if no errors
        """
        red = "background-color: rgba(255, 0, 0, 128);"
        success = True

        # Check 1: k_size
        if self.k_size == "":
            self.k_size.setStyleSheet(red)
            success = False
        else:
            try:
                if self.current_kernel_type == "median":
                    k_size = int(self.k_size.text())
                else:
                    k_size = float(self.k_size.text())
                if k_size <= 0:
                    self.k_size.setStyleSheet(red)
                    success = False
                else:
                    self.k_size.setStyleSheet("")
            except ValueError:
                if self.current_kernel_type == "median":
                    info = QMessageBox.critical(self, "Error",
                                                "Kernel size must be integer for median")
                self.k_size.setStyleSheet(red)
                success = False

        return success

    def call_main(self):
        try:
            self.main()
        except Exception as e:
            info = QMessageBox.critical(self, "Error", str(e))
            self.cancel()
            raise

    def main(self):
        """
        Main function to process input and call smoothing function
        """
        success = self.input_validation()

        if not success:
            return

        self.hide()
        self.abort_window = AbortWindow(self)
        QApplication.processEvents()

        # Add smoothing parameters

        self.smooth_cube.abort_window = self.abort_window
        if self.smooth_cube.parent is None and self.parent is not self.smooth_cube:
            self.smooth_cube.parent = self.parent
        if self.parent is not self.smooth_cube:
            self.smooth_cube.data = self.data
        self.smooth_cube.smoothing_axis = self.current_axis
        self.smooth_cube.kernel_type = self.current_kernel_type
        if self.current_kernel_type == "median":
            self.smooth_cube.kernel_size = int(self.k_size.text())
        else:
            self.smooth_cube.kernel_size = float(self.k_size.text())
        self.smooth_cube.component_id = str(self.component_combo.currentText())
        self.smooth_cube.output_as_component = True

        if self.is_preview_active:
            self.parent.end_smoothing_preview()
            self.is_preview_active = False
        self.smooth_cube.multi_threading_smooth()
        return

    def update_preview_button(self):
        if self.parent is None or "spatial" != self.current_axis:
            self.previewButton.setDisabled(True)
            return
        self.previewButton.setDisabled(False)
        return

    def call_preview(self):
        try:
            self.preview()
        except Exception as e:
            info = QMessageBox.critical(self, "Error", str(e))
            self.cancel()
            raise

    def preview(self):
        """Preview current options"""
        success = self.input_validation()

        if not success:
            return

        if self.smooth_cube.parent is None and self.parent is not self.smooth_cube:
            self.smooth_cube.parent = self.parent
        self.smooth_cube.smoothing_axis = self.current_axis
        self.smooth_cube.kernel_type = self.current_kernel_type
        if self.current_kernel_type == "median":
            self.smooth_cube.kernel_size = int(self.k_size.text())
        else:
            self.smooth_cube.kernel_size = float(self.k_size.text())

        preview_function = self.smooth_cube.preview_smoothing
        preview_title = self.smooth_cube.get_preview_title()
        component_id = self.component_combo.currentData()
        self.parent.start_smoothing_preview(preview_function, component_id, preview_title)

        self.is_preview_active = True
        self.preview_message.show()

    def cancel(self):
        self.clean_up()

    def clean_up(self):
        self.close()
        if self.abort_window is not None:
            self.abort_window.close()
        if self.is_preview_active:
            self.parent.end_smoothing_preview()
            self.is_preview_active = False

    def closeEvent(self, event):
        if self.is_preview_active:
            self.parent.end_smoothing_preview()
            self.is_preview_active = False

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.clean_up()



from astropy import units as u
from astropy.units.quantity import Quantity


class CustomFluxEquivalences:
    """
    This class is intended to behave as a function.
    It will replace astropy.units.equivalencies.spectral_density.
    It saves the original spectral_destiny function and adds
    flux units that are over pixels or arcsec**2. The class also
    stores pixel_area information that is used to convert b/w the
    pixels and arcsec**2.
    """
    def __init__(self, spectral_density):
        self.pixel_area = None
        self.default_spectral_density = spectral_density

    def __call__(self, wave, factor=None):
        pixel_area = self.pixel_area
        if pixel_area is not None:
            if type(pixel_area) == Quantity:
                pixel_area = pixel_area.to("arcsec2 / pix").value  # Convert to Quantity
        default_spectral_density = self.default_spectral_density(wave, factor=None)
        equivalencies = default_spectral_density[:]
        added_area_units = []
        for u1, u2, f1, f2 in default_spectral_density:
            u1_pix = u1 / u.pix
            u2_pix = u2 / u.pix

            u1_area = u1 / (u.arcsec ** 2)
            u2_area = u2 / (u.arcsec ** 2)

            equivalencies.append((u1_pix, u2_pix, f1, f2))
            equivalencies.append((u1_area, u2_area, f1, f2))

            if pixel_area is not None:
                equivalencies.append((u1_pix,
                                      u2_area,
                                      lambda x: f1(x) / pixel_area,
                                      lambda x: f2(x) * pixel_area))
                equivalencies.append((u1_area,
                                      u2_pix,
                                      lambda x: f1(x) * pixel_area,
                                      lambda x: f2(x) / pixel_area))
                if u1_area not in added_area_units:
                    equivalencies.append((u1_area,
                                          u1_pix,
                                          lambda x: x * pixel_area,
                                          lambda x: x / pixel_area))
                    added_area_units.append(u1_area)

                if u2_area not in added_area_units:
                    equivalencies.append((u2_area,
                                          u2_pix,
                                          lambda x: x * pixel_area,
                                          lambda x: x / pixel_area))
                    added_area_units.append(u2_area)

        return equivalencies



import os
import matplotlib.cm as cm

from qtpy.QtWidgets import (QLabel, QAction, QActionGroup,
                            QDialog, QHBoxLayout, QVBoxLayout,
                            QPushButton, QCheckBox, QLineEdit,
                            QGroupBox, QGridLayout, QMessageBox)
from qtpy.QtCore import Qt

from glue.config import viewer_tool
from glue.config import colormaps as glue_colormaps
from glue.utils.qt import QColormapCombo

from glue.viewers.common.qt.tool import Tool

DEFAULT_GLUE_COLORMAP_INDEX = 3
DEFAULT_CONTOUR_FONT_SIZE = 10
ICON_PATH = os.path.abspath(
    os.path.join(
        os.path.abspath(__file__),
        "..",
        "..",
        "data",
        "resources",
        "contour_icon.png"
    )
)

@viewer_tool
class ContourButton(Tool):
    """
    Contour map tool bar menu and button
    """
    icon = ICON_PATH
    tool_id = 'cubeviz:contour'
    action_text = 'Toggles contour map'
    tool_tip = 'Toggles contour map'
    status_tip = ''
    shortcut = None

    def __init__(self, viewer):
        super(ContourButton, self).__init__(viewer)
        self.tool_bar = self.viewer.toolbar
        if self.viewer.cubeviz_layout is not None:
            self.cubeviz_layout = self.viewer.cubeviz_layout
        else:
            self.cubeviz_layout = None

        self.options = None

    def activate(self):
        pass

    def menu_actions(self):
        """
        List of QtWidgets.QActions to be attached to this tool
        as a context menu.
        """
        self.options = []

        # WARNING: QAction labels are used to identify them.
        #          Changing them can cause problems unless
        #          all references are updated in this package.
        component_action_group = QActionGroup(self.tool_bar)

        action = QAction("Off", self.tool_bar, checkable=True)
        action.setChecked(True)
        action.setActionGroup(component_action_group)
        action.triggered.connect(self.viewer.remove_contour)
        self.options.append(action)

        action = QAction("Current Component", self.tool_bar, checkable=True)
        action.setActionGroup(component_action_group)
        action.triggered.connect(self.viewer.default_contour)
        self.options.append(action)

        action = QAction("Other Component", self.tool_bar, checkable=True)
        action.setActionGroup(component_action_group)
        action.triggered.connect(self.viewer.custom_contour)
        self.options.append(action)

        action = QAction(" ", self.tool_bar)
        action.setSeparator(True)
        self.options.append(action)

        action = QAction("Contour Settings", self.tool_bar)
        action.triggered.connect(self.viewer.edit_contour_settings)
        self.options.append(action)

        return self.options


class ContourOptionsDialog(QDialog):
    """
    Dialog box for selecting contour options
    """
    def __init__(self, contour_settings):
        super(ContourOptionsDialog, self).__init__(contour_settings.cubeviz_layout)
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        self.setWindowTitle("Contour Options")

        self.is_preview_active = False  # preview mode?

        self.contour_settings = contour_settings  # ref to caller ContourSettings
        self.image_viewer = self.contour_settings.image_viewer  # ref to image viewer
        self.options = self.contour_settings.options  # ref to ContourSettings options

        self._colormap_members = self.contour_settings.colormap_members  # Colormap options
        self._colormap_index = DEFAULT_GLUE_COLORMAP_INDEX  # Currently selected colormap
        if "cmap" in self.options:
            if self.options["cmap"] in self._colormap_members:
                self._colormap_index = self._colormap_members.index(self.options["cmap"])

        # Is there a user spacing?
        if self.contour_settings.spacing is None:
            self.is_custom_spacing = False
        else:
            self.is_custom_spacing = True
        # Is there a user min?
        if self.contour_settings.vmin is None:
            self.is_vmin = False
        else:
            self.is_vmin = True
        # Is there a user max?
        if self.contour_settings.vmax is None:
            self.is_vmax = False
        else:
            self.is_vmax = True

        self.add_contour_label = self.contour_settings.add_contour_label  # bool

        self._init_ui()

    def _init_ui(self):

        # Line 1: Color map
        self.colormap_label = QLabel("Color Scheme: ")

        self.colormap_combo = QColormapCombo()
        self.colormap_combo.addItem("", userData=cm.viridis)
        self.colormap_combo._update_icons()
        self.colormap_combo.setCurrentIndex(self._colormap_index)
        self.colormap_combo.setMaximumWidth(150)
        self.colormap_combo.currentIndexChanged.connect(
            self._on_colormap_change)

        #   hbl is short for Horizontal Box Layout
        hbl1 = QHBoxLayout()
        hbl1.addWidget(self.colormap_label)
        hbl1.addWidget(self.colormap_combo)

        # Line 2: Display contour labels
        self.contour_label_checkBox = QCheckBox("Contour labels (font size):")
        if self.contour_settings.add_contour_label:
            self.contour_label_checkBox.setChecked(True)
        self.contour_label_checkBox.toggled.connect(self.toggle_labels)

        font_string = str(self.contour_settings.font_size)
        self.font_size_input = QLineEdit(font_string)
        self.font_size_input.setFixedWidth(150)
        self.font_size_input.setDisabled(
            not self.contour_settings.add_contour_label)

        hbl2 = QHBoxLayout()
        hbl2.addWidget(self.contour_label_checkBox)
        hbl2.addWidget(self.font_size_input)

        # Line 3: Contour Spacing
        self.custom_spacing_checkBox = QCheckBox("Contour spacing (interval):")
        if self.is_custom_spacing:
            self.custom_spacing_checkBox.setChecked(True)
        self.custom_spacing_checkBox.toggled.connect(self.custom_spacing)

        self.spacing_input = QLineEdit()
        self.spacing_input.setFixedWidth(150)
        self.spacing_input.setDisabled(not self.is_custom_spacing)
        spacing = ""
        if self.is_custom_spacing:
            spacing = str(self.contour_settings.spacing)
        elif self.contour_settings.data_spacing is not None:
            spacing = self.contour_settings.data_spacing
            spacing = "{0:1.4f}".format(spacing)
        self.spacing_default_text = spacing
        self.spacing_input.setText(spacing)

        hbl3 = QHBoxLayout()
        hbl3.addWidget(self.custom_spacing_checkBox)
        hbl3.addWidget(self.spacing_input)

        # Line 4: Vmax
        self.vmax_checkBox = QCheckBox("Set max:")

        self.vmax_input = QLineEdit()
        self.vmax_input.setFixedWidth(150)
        self.vmax_input.setDisabled(not self.is_vmax)

        vmax = ""
        if self.is_vmax:
            self.vmax_checkBox.setChecked(True)
            vmax = str(self.contour_settings.vmax)
        elif self.contour_settings.data_max is not None:
            vmax = self.contour_settings.data_max
            vmax = "{0:1.4f}".format(vmax)
        self.vmax_input.setText(vmax)
        self.vmax_default_text = vmax

        self.vmax_checkBox.toggled.connect(self.toggle_vmax)

        hbl4 = QHBoxLayout()
        hbl4.addWidget(self.vmax_checkBox)
        hbl4.addWidget(self.vmax_input)

        # Line 5: Vmin
        self.vmin_checkBox = QCheckBox("Set min:")

        self.vmin_input = QLineEdit()
        self.vmin_input.setFixedWidth(150)
        self.vmin_input.setDisabled(not self.is_vmin)

        vmin = ""
        if self.is_vmin:
            self.vmin_checkBox.setChecked(True)
            vmin = str(self.contour_settings.vmin)
        elif self.contour_settings.data_min is not None:
            vmin = self.contour_settings.data_min
            vmin = "{0:1.4f}".format(vmin)
        self.vmin_input.setText(vmin)
        self.vmin_default_text = vmin

        self.vmin_checkBox.toggled.connect(self.toggle_vmin)

        hbl5 = QHBoxLayout()
        hbl5.addWidget(self.vmin_checkBox)
        hbl5.addWidget(self.vmin_input)

        # Line f:
        self.previewButton = QPushButton("Preview")
        self.previewButton.clicked.connect(self.preview)

        self.defaultButton = QPushButton("Reset")
        self.defaultButton.clicked.connect(self.default)

        self.okButton = QPushButton("OK")
        self.okButton.clicked.connect(self.finish)
        self.okButton.setDefault(True)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.cancel)

        hblf = QHBoxLayout()
        hblf.addStretch(1)
        hblf.addWidget(self.previewButton)
        hblf.addWidget(self.defaultButton)
        hblf.addWidget(self.cancelButton)
        hblf.addWidget(self.okButton)

        vbl = QVBoxLayout()
        vbl.addLayout(hbl1)
        vbl.addLayout(hbl2)
        vbl.addLayout(hbl3)
        vbl.addLayout(hbl4)
        vbl.addLayout(hbl5)
        vbl.addLayout(hblf)

        self.setLayout(vbl)

        self.show()

    def update_data_vals(self, vmin="", vmax="", spacing="1"):

        self.vmin_default_text = vmin

        if not self.is_vmin:
            self.vmin_input.setText(vmin)

        self.vmax_default_text = vmax
        if not self.is_vmax:
            self.vmax_input.setText(vmax)

        self.spacing_default_text = spacing
        if not self.is_custom_spacing:
            self.spacing_input.setText(spacing)

    def _on_colormap_change(self, index):
        """Combo index changed handler"""
        self._colormap_index = index

    def custom_spacing(self):
        """Checkbox toggled handler"""
        if self.is_custom_spacing:
            self.is_custom_spacing = False
            self.spacing_input.setDisabled(True)
            spacing = ""
            if self.contour_settings.data_spacing:
                spacing = self.contour_settings.data_spacing
                spacing = "{0:1.4f}".format(spacing)
            self.spacing_input.setText(spacing)
            self.spacing_input.setStyleSheet("")
        else:
            self.is_custom_spacing = True
            self.spacing_input.setDisabled(False)

    def toggle_labels(self):
        """Checkbox toggled handler"""
        if self.add_contour_label:
            self.add_contour_label = False
            self.font_size_input.setDisabled(True)
            font_string = str(self.contour_settings.font_size)
            self.font_size_input.setText(font_string)
            self.font_size_input.setStyleSheet("")
        else:
            self.add_contour_label = True
            self.font_size_input.setDisabled(False)

    def toggle_vmax(self):
        """Checkbox toggled handler"""
        if self.is_vmax:
            self.is_vmax = False
            self.vmax_input.setDisabled(True)
            vmax = ""
            if self.contour_settings.data_max:
                vmax = self.contour_settings.data_max
                vmax = "{0:1.4f}".format(vmax)
            self.vmax_input.setText(vmax)
            self.vmax_input.setStyleSheet("")
        else:
            self.is_vmax = True
            self.vmax_input.setDisabled(False)

    def toggle_vmin(self):
        """Checkbox toggled handler"""
        if self.is_vmin:
            self.is_vmin = False
            self.vmin_input.setDisabled(True)
            vmin = ""
            if self.contour_settings.data_min:
                vmin = self.contour_settings.data_min
                vmin = "{0:1.4f}".format(vmin)
            self.vmin_input.setText(vmin)
            self.vmin_input.setStyleSheet("")
        else:
            self.is_vmin = True
            self.vmin_input.setDisabled(False)

    def input_validation(self):
        red = "background-color: rgba(255, 0, 0, 128);"

        def float_check(min_val=None):
            if user_input.text() == "":
                user_input.setStyleSheet(red)
                return False
            else:
                try:
                    value = float(user_input.text())
                    if min_val is not None:
                        if value <= min_val:
                            user_input.setStyleSheet(red)
                            return False
                    else:
                        user_input.setStyleSheet("")
                except ValueError:
                    user_input.setStyleSheet(red)
                    return False
            return True

        def int_check(min_val=None):
            if user_input.text() == "":
                user_input.setStyleSheet(red)
                return False
            else:
                try:
                    value = int(user_input.text())
                    if min_val is not None:
                        if value <= min_val:
                            user_input.setStyleSheet(red)
                            return False
                    else:
                        user_input.setStyleSheet("")
                except ValueError:
                    user_input.setStyleSheet(red)
                    return False
            return True

        success = True

        # Check 1: spacing_input
        if self.is_custom_spacing:
            user_input = self.spacing_input
            float_check(0)
            success = success and float_check()

        # Check 2: font_size_input
        if self.add_contour_label:
            user_input = self.font_size_input
            int_check(0)
            success = success and int_check()

        # Check 3: vmax
        if self.is_vmax:
            user_input = self.vmax_input
            float_check()
            success = success and float_check()

        # Check 4: vmax
        if self.is_vmin:
            user_input = self.vmin_input
            float_check()
            success = success and float_check()

        # Check 5: vmax and vmin
        if self.is_vmax and self.is_vmin and success:
            vmax = float(self.vmax_input.text())
            vmin = float(self.vmin_input.text())
            if vmax <= vmin:
                self.vmax_input.setStyleSheet(red)
                self.vmin_input.setStyleSheet(red)
                success = False

        return success

    def finish(self):
        """
        Ok button pressed. Finalize
        options and send to image viewer
         """
        success = self.input_validation()

        if not success:
            return

        # Change Color Map
        self._colormap_index = self.colormap_combo.currentIndex()
        colormap = self._colormap_members[self._colormap_index]
        self.contour_settings.options["cmap"] = colormap

        # labels
        self.contour_settings.add_contour_label = self.add_contour_label

        # font size
        if self.add_contour_label:
            font_size = int(self.font_size_input.text())
            self.contour_settings.font_size = font_size
        else:
            self.contour_settings.font_size = DEFAULT_CONTOUR_FONT_SIZE

        # Spacing
        if self.is_custom_spacing:
            self.contour_settings.spacing = float(self.spacing_input.text())
        else:
            self.contour_settings.spacing = None

        # vmax
        if self.is_vmax:
            vmax = float(self.vmax_input.text())
            self.contour_settings.vmax = vmax
            self.contour_settings.options["vmax"] = vmax
        else:
            self.contour_settings.vmax = None
            self.contour_settings.options["vmax"] = None

        # vmin
        if self.is_vmin:
            vmin = float(self.vmin_input.text())
            self.contour_settings.vmin = vmin
            self.contour_settings.options["vmin"] = vmin
        else:
            self.contour_settings.vmin = None
            self.contour_settings.options["vmin"] = None

        # Redraw contour
        if self.contour_settings.image_viewer.is_contour_active:
            self.contour_settings.draw_function()

        self.close()

    def preview(self):
        """
        Prepare preview contour settings
        and send to image viewer
        """
        success = self.input_validation()

        if not success:
            return

        image_viewer = self.contour_settings.image_viewer
        preview_settings = ContourSettings(image_viewer)
        preview_settings.dialog = self
        preview_settings.options = self.contour_settings.options.copy()
        preview_settings.spacing = self.contour_settings.spacing

        # Change Color Map
        self._colormap_index = self.colormap_combo.currentIndex()
        colormap = self._colormap_members[self._colormap_index]
        preview_settings.options["cmap"] = colormap

        # labels
        add_contour_label = self.contour_label_checkBox.isChecked()
        preview_settings.add_contour_label = add_contour_label

        # font size
        if add_contour_label:
            font_size = int(self.font_size_input.text())
            preview_settings.font_size = font_size

        # Spacing
        if self.is_custom_spacing:
            preview_settings.spacing = float(self.spacing_input.text())
        else:
            preview_settings.spacing = None

        # vmax
        if self.is_vmax:
            vmax = float(self.vmax_input.text())
            preview_settings.vmax = vmax
            preview_settings.options["vmax"] = vmax
        else:
            preview_settings.vmax = None
            preview_settings.options["vmax"] = None

        # vmin
        if self.is_vmin:
            vmin = float(self.vmin_input.text())
            preview_settings.vmin = vmin
            preview_settings.options["vmin"] = vmin
        else:
            preview_settings.vmin = None
            preview_settings.options["vmin"] = None

        # Redraw contour
        if image_viewer.is_contour_active:
            self.is_preview_active = True
            image_viewer.set_contour_preview(preview_settings)
        else:
            message = "Contour map is currently switched off. " \
                      "Please turn on the contour map by selecting " \
                      "a component from the contour map drop-down menu."
            info = QMessageBox.critical(self, "Error", message)

    def default(self):
        """
        Set options back to default
        and send to image viewer
        """
        self.contour_settings.options = self.contour_settings.default_options()
        self.contour_settings.spacing = None
        self.contour_settings.font_size = DEFAULT_CONTOUR_FONT_SIZE
        self.contour_settings.vmax = None
        self.contour_settings.vmin = None
        self.contour_settings.add_contour_label = False
        if self.contour_settings.image_viewer.is_contour_active:
            self.contour_settings.draw_function()
        self.contour_settings.options_dialog()

    def cancel(self):
        if self.contour_settings.image_viewer.is_contour_active:
            self.contour_settings.draw_function()
        self.close()

    def closeEvent(self, event):
        """closeEvent handler"""
        if self.is_preview_active:
            self.contour_settings.image_viewer.end_contour_preview()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Escape:
            self.cancel()


class ContourSettings(object):
    """
    A class that contains contour settings.
    Settings
    ---------
    options : dict
        Holds kwargs to matplotlib.axes.Axes.contour
    spacing : int
        Spacing between contours
    vmax and vmin: float and float
        max and min values of the contour
    """
    def __init__(self, image_viewer, data_max=None,
                 data_min=None, data_spacing=None):
        self.image_viewer = image_viewer
        self.cubeviz_layout = self.image_viewer.cubeviz_layout
        self.draw_function = self.image_viewer.draw_contour

        self.options = self.default_options()

        self.spacing = None  # contour spacing

        self.add_contour_label = False  # add valye labels?
        self.font_size = DEFAULT_CONTOUR_FONT_SIZE

        self.vmax = None # Max value of contour
        self.vmin = None # Min value of contour

        # Load colormap options
        #   Add viridis to drop down
        self.colormap_members = [m[1] for m in glue_colormaps.members]
        self.colormap_members.append(cm.viridis)

        # Default data values
        self.data_max = data_max
        self.data_min = data_min
        self.data_spacing = data_spacing

        # Dialog
        self.dialog = None

    @staticmethod
    def default_options():
        options = {"cmap": cm.viridis}

        return options

    def options_dialog(self):
        """
        Open UI to change settings
        :return: qtpy.QtWidgets.QDialog
        """
        if self.dialog is not None:
            self.dialog.close()
        self.dialog = ContourOptionsDialog(self)
        return self.dialog

    def update_dialog(self):
        vmin = ""
        if self.data_min is not None:
            vmin = "{0:1.4f}".format(self.data_min)
        vmax = ""
        if self.data_max is not None:
            vmax = "{0:1.4f}".format(self.data_max)

        spacing = ""
        if self.data_spacing is not None:
            spacing = "{0:1.4f}".format(self.data_spacing)

        self.dialog.update_data_vals(vmin, vmax, spacing)

    def is_simple_contour(self):
        """
        Check to see if settings are simple enough
        to use `contour(arr, **settings.options)?`
        """
        is_simple = True
        if self.vmax is not None:
            is_simple = False
        elif self.vmin is not None:
            is_simple = False
        elif self.spacing is not None:
            is_simple = False
        return is_simple



import os

from glue.core.data_factories import find_factory

from ..data_factories import DataFactoryConfiguration, cubeviz_fits_exporter

TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data', 'data_cube.fits.gz')


def test_export_roundtrip(tmpdir):

    # TODO: generalize this to all example data files once we
    # have more than one format in the data directory.

    DataFactoryConfiguration(check_ifu_valid=False)

    # Make sure the right factory was identified
    factory = find_factory(TEST_DATA_PATH)
    assert factory.__self__.name == 'kmos'

    # Load in the data
    data = factory(TEST_DATA_PATH)

    # Generate temporary output filename
    filename = tmpdir.join('test.fits').strpath

    # Export the data
    cubeviz_fits_exporter(filename, data)

    # And check that we identify the file correctly again
    factory = find_factory(filename)
    assert factory.__self__.name == 'kmos'



from qtpy import QtWidgets


class CubevizToolbar(QtWidgets.QToolBar):

    def __init__(self, application=None, parent=None):

        super(CubevizToolbar, self).__init__(parent=parent)

        self.application = application

        self._button_viewer_options = QtWidgets.QPushButton()
        self._button_viewer_options.setText("Hide viewer options")
        self._button_viewer_options.clicked.connect(self._toggle_sidebar)

        self.addWidget(self._button_viewer_options)

    def _toggle_sidebar(self, event=None):
        splitter = self.application._ui.main_splitter
        sizes = list(splitter.sizes())
        if sizes[0] == 0:
            sizes[0] += 10
            sizes[1] -= 10
            self._button_viewer_options.setText("Hide viewer options")
        else:
            sizes[1] = sizes[0] + sizes[1]
            sizes[0] = 0
            self._button_viewer_options.setText("Show viewer options")
        splitter.setSizes(sizes)



from astropy import units as u
from specviz.third_party.glue.data_viewer import dispatch as specviz_dispatch

from ..messages import WavelengthUpdateMessage, WavelengthUnitUpdateMessage, RedshiftUpdateMessage


OBS_WAVELENGTH_TEXT = 'Obs Wavelength'
REST_WAVELENGTH_TEXT = 'Rest Wavelength'


class WavelengthController:
    def __init__(self, cubeviz_layout):
        self._cv_layout = cubeviz_layout
        self._hub = cubeviz_layout.session.hub
        ui = cubeviz_layout.ui
        self._wavelengths = []
        self._original_wavelengths = []
        self._original_units = u.m
        self._current_units = self._original_units

        # Add the redshift z value
        self._redshift_z = 0

        # This is the Wavelength conversion/combobox code
        self._units = [u.m, u.cm, u.mm, u.um, u.nm, u.AA]
        self._units_titles = list(u.long_names[0].title() for u in self._units)

        # This is the label for the wavelength units
        self._wavelength_textbox_label = ui.wavelength_textbox_label.text()

        specviz_dispatch.setup(self)

    def enable(self, units, wavelength):
        self._wavelengths = wavelength
        self._original_wavelengths = wavelength
        self._send_wavelength_message(wavelength)
        self._send_wavelength_unit_message(units)

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def wavelength_label(self):
        return self._wavelength_textbox_label

    @wavelength_label.setter
    def wavelength_label(self, value):
        self._wavelength_textbox_label = value

    @property
    def units(self):
        return self._units

    @property
    def current_units(self):
        return self._current_units

    @property
    def unit_titles(self):
        return self._units_titles

    @property
    def redshift_z(self):
        """
        Get the Z value for the redshift.

        :return: Z redshift value
        """
        return self._redshift_z

    @specviz_dispatch.register_listener("change_redshift")
    def specviz_change_redshift(self, redshift):
        """
        Change the redshift based on a message from specviz.

        :paramter: redshift - the Z value.
        :return: nothing
        """
        self.update_redshift(redshift)

    def update_units(self, units):

        self._wavelengths = (self._wavelengths * self._current_units).to(units) / units

        self._current_units = units

        self._send_wavelength_unit_message(units)
        self._send_wavelength_message(self._wavelengths)

        specviz_dispatch.changed_units.emit(x=units)

    def update_redshift(self, redshift, label=''):
        # If the input redshift is the current value we have then we are not
        # going to do anything.
        if self._redshift_z == redshift:
            return

        if redshift is not None and redshift != 0:
            self._wavelength_textbox_label = REST_WAVELENGTH_TEXT
        else:
            self._wavelength_textbox_label = OBS_WAVELENGTH_TEXT

        orig_redshift = (1.0 / (1 + redshift)) * self._original_wavelengths * self._original_units
        self._wavelengths = orig_redshift.to(self._current_units)/ self._current_units

        # This calls the setter above, so really, the magic is there.
        self._redshift_z = redshift

        self._send_redshift_message(redshift)
        self._send_wavelength_message(self._wavelengths)

        specviz_dispatch.change_redshift.emit(redshift=redshift)

    def _send_wavelength_message(self, wavelengths):
        msg = WavelengthUpdateMessage(self, wavelengths)
        self._hub.broadcast(msg)

    def _send_wavelength_unit_message(self, units):
        msg = WavelengthUnitUpdateMessage(self, units)
        self._hub.broadcast(msg)

    def _send_redshift_message(self, redshift):
        msg = RedshiftUpdateMessage(self, redshift, label=self.wavelength_label)
        self._hub.broadcast(msg)



from glue.config import keyboard_shortcut
from qtpy import QtCore, QtWidgets, QtGui, compat
from qtpy.QtWidgets import QApplication
from glue.config import viewer_tool
import matplotlib.pyplot as plt

from .tools.wavelengths_ui import WavelengthUI

@keyboard_shortcut(QtCore.Qt.Key_A, None)
def move_slider_left(session):
    """
    Move slider index one to the left
    :param session:
    :return:
    """
    curr_layout = session.application.current_tab.ui
    curr_layout.change_slice_index(-1)


@keyboard_shortcut(QtCore.Qt.Key_D, None)
def move_slider_right(session):
    """
    Move slider index one to the right
    :param session:
    :return:
    """
    curr_layout = session.application.current_tab.ui
    curr_layout.change_slice_index(1)


@keyboard_shortcut(QtCore.Qt.Key_F, None)
def lock_coordinates(session):
    """
    Lock coordinates in the bottom right of the viewer window
    :param session:
    :return:
    """
    lay = session.application.current_tab
    for w in lay.cube_views:
        civ = w._widget
        if civ.is_mouse_over:
            civ.toggle_hold_coords()


@keyboard_shortcut(QtCore.Qt.Key_P, None)
def copy_coordinates_to_clipboard(session):
    """
    Copy coordinates to clipboard so user can paste them elsewhere
    :param session:
    :return:
    """
    coords_status = False
    lay = session.application.current_tab
    for w in lay.cube_views:
        civ = w._widget
        if civ.is_mouse_over:
            coords_status = civ.get_coords()
            break

    if coords_status:
        cb = QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(coords_status, mode=cb.Clipboard)


@keyboard_shortcut(QtCore.Qt.Key_1, None)
def show_wavelength_dialog(session):
    """
    Popup the Wavelength dialog in order to change the units or redshift.

    :param session:
    :return:
    """
    WavelengthUI(session.application.tab(0)._wavelength_controller, parent=session.application.tab(0))


def remove_mpl_shortcuts_and_check_dupes():
    """
    Makes sure that shortcut keys do not overlap between the three things (glue.config.viewer_tool, MatPlotLib,
    and keyboard_shortcuts) that use them
    :return:
    """
    # Get all shortcuts from glue.config.viewer_tools
    vt_shortcuts = []
    for vt in viewer_tool.members:
        if viewer_tool.members[vt].shortcut is not None:
            vt_shortcuts.append(viewer_tool.members[vt].shortcut)

    # If shortcut already exists in glue.config.viewer_tool, raise error
    ks_shortcuts = []
    for ks in keyboard_shortcut.members[None]:
        if ks < 128 and chr(ks) in vt_shortcuts:
            print("Make sure the shortcuts are NOT one of the following {0}\n\n".format(vt_shortcuts))
            print("Keyboard shortcut '{0}' already registered in {1}".format(chr(ks), "glue.config.viewer_tool"))
        elif ks < 128:
            ks_shortcuts.append(chr(ks))
        elif ks >= 128:
            print("Please refrain from using non-ASCII values for shortcuts")

    # Remove MatPlotLib default shortcuts if they conflict with other shortcuts
    for param in plt.rcParams:
        if 'keymap' in param:
            for key in plt.rcParams[param]:
                if key.upper() in vt_shortcuts or key.upper() in ks_shortcuts:
                    plt.rcParams[param].remove(key)


remove_mpl_shortcuts_and_check_dupes()



import numpy as np
from glue.core.data import Data
from glue.config import colormaps as glue_colormaps


DEFAULT_GLUE_COLORMAP_INDEX = 3


class OverlayController:

    def __init__(self, cubeviz_layout):
        self._cv_layout = cubeviz_layout
        self._cube_views = cubeviz_layout.cube_views
        ui = cubeviz_layout.ui

        self._overlays = Data('Overlays')
        # This is a list of overlay objects that are currently displayed
        self._active_overlays = []
        # Maps overlays to the data sets they represent
        self._overlay_map = {}
        self._overlay_colorbar_axis = []

        self._overlay_image_combo = ui.overlay_image_combo
        self._overlay_colormap_combo = ui.overlay_colormap_combo
        self._alpha_slider = ui.alpha_slider
        self._colormap_index = DEFAULT_GLUE_COLORMAP_INDEX

        self._overlay_image_combo.addItem("No Overlay")
        self._overlay_image_combo.currentIndexChanged.connect(
            self._on_overlay_change)

        self._overlay_colormap_combo.setCurrentIndex(self._colormap_index)
        self._overlay_colormap_combo.currentIndexChanged.connect(
            self._on_colormap_change)

        self._alpha_slider.valueChanged.connect(self._on_alpha_change)

    def add_overlay(self, data, label, display=True):
        self._overlays.add_component(data, label)
        # TODO: Is there a way to get this from the component ???
        self._overlay_image_combo.addItem(label)
        new_index = self._overlay_image_combo.count() - 1
        self._overlay_map[new_index] = data

        self._alpha_slider.setEnabled(True)
        self._overlay_image_combo.setEnabled(True)
        self._overlay_colormap_combo.setEnabled(True)

        if display:
            # Setting the index will cause _on_overlay_change to fire
            self._overlay_image_combo.setCurrentIndex(new_index)

    def _on_overlay_change(self, index):
        if index == 0:
            data = None
        else:
            data = self._overlay_map[index]
        self.display_overlay(data)

    def _on_colormap_change(self, index):
        self._colormap_index = index
        colormap = glue_colormaps.members[self._colormap_index][1]
        for overlay in self._active_overlays:
            overlay.set_cmap(colormap)
        for cb in self._overlay_colorbar_axis:
            for cbim in cb.get_images():
                cbim.set_cmap(colormap)
        for cube in self._cube_views:
            cube._widget.figure.canvas.draw()

    def _draw_mpl_overlay(self, data, view):
        axes = view._widget.axes
        aspect = axes.get_aspect()

        extent = 0, data.shape[0], 0, data.shape[1]

        colormap = glue_colormaps.members[self._colormap_index][1]
        overlay = view._widget.axes.imshow(
            data, origin='lower', cmap=colormap, alpha=.25,
            interpolation='none', aspect=aspect, extent=extent)

        self._active_overlays.append(overlay)

        # Add the overlay colorbar as an axis
        oca = view._widget.figure.add_axes([0.02, 0.04, 0.3, 0.025], projection='rectilinear')
        mindata, maxdata = np.nanmin(data), np.nanmax(data)
        oca_image = np.zeros((1,100))
        oca_image[0] = np.arange(mindata, maxdata, (maxdata-mindata)/100)
        oca.imshow(oca_image, origin='lower', cmap=colormap, aspect=aspect, extent=[0,100,0,100])
        oca.set_xticks([0, 25, 50, 75, 100])
        oca.set_xticklabels(['%3.2e'%x for x in np.arange(mindata, maxdata, (maxdata-mindata)/5)], fontsize=6)
        oca.set_yticks([])
        self._overlay_colorbar_axis.append(oca)

        view._widget.figure.canvas.draw()

    def display_overlay(self, data):
        # Remove all existing overlays
        if self._active_overlays:
            for overlay, view, cb in zip(
                    self._active_overlays, self._cube_views, self._overlay_colorbar_axis):
                overlay.remove()
                cb.remove()
                view._widget.figure.canvas.draw()

            self._active_overlays = []
            self._overlay_colorbar_axis = []

        # Just return if no new overlay is to be drawn
        if data is None:
            return

        self._active_overlays = []
        for view in self._cube_views:
            self._draw_mpl_overlay(data, view)

        self._alpha_slider.setValue(25)

    def _on_alpha_change(self, event):
        """
        Callback for change in alpha value.

        :param event:
        :return:
        """
        for overlay in self._active_overlays:
            overlay.set_alpha(self._alpha_slider.value() / 100.)
            overlay.figure.canvas.draw()



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2017-10-26 09:59:52.871866
from __future__ import unicode_literals
import datetime

version = "2.0"
githash = "ed2e7897862ae9e979b336df670d7a5541cdd8f0"


major = 2
minor = 0
bugfix = 0

release = True
timestamp = datetime.datetime(2017, 10, 26, 9, 59, 52, 871866)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



import os
import logging
from collections import defaultdict

from astropy.io import fits
from astropy import units as u

logging.basicConfig(level=logging.DEBUG, format="%(filename)s: %(levelname)8s %(message)s")
log = logging.getLogger('ifcube')
log.setLevel(logging.WARNING)


class IFUCube(object):
    """
    Check and correct the IFUCube
    """

    def __init__(self):
        self._fits = None
        self._filename = None
        self._good = True
        self._log_text = defaultdict(lambda: dict())

        self._units = [u.m, u.cm, u.mm, u.um, u.nm, u.AA]
        self._units_titles = list(x.name for x in self._units)

    def open(self, filename, fix=False):
        """
        Check all checkers
        """
        self._filename = filename

        log.debug('In check with filename {} and fix {}'.format(filename, fix))

        # Check existence of the file
        if not os.path.isfile(filename):
            log.warning('File {} does not exist'.format(filename))
            return
        else:
            log.info('File {} exists'.format(filename))

        # Open the file
        try:
            self._fits = fits.open(filename)
        except:
            log.warning('Could not open {} '.format(filename))
            return

        self.check(fix)

        return self._fits

    def check(self, fix=False):
        """
        Check all checkers
        """
        log.debug('In check with filename {} and fix {}'.format(self._filename, fix))
        self._log_text['>front'] = 'Checking filename {}\n'.format(self._filename)

        self.check_data(fix)

        self.check_ctype1(fix)

        self.check_ctype2(fix)

        self.check_ctype3(fix)

        self.check_cunit1(fix)

        self.check_cunit2(fix)

        self.check_cunit3(fix)

        return self._fits

    def check_data(self, fix=False):
        """
        Check CTYPE and make sure it is the correct value

        :param: fits_file: The open fits file
        :param: fix: boolean whether to fix it or not
        :return: boolean whether it is good or not
        """
        log.debug('In check_data')
        good = False
        data_shape = []

        for ii, hdu in enumerate(self._fits):

            # Check the EXTNAME field for this HDU
            if not 'EXTNAME' in hdu.header:
                log.warning(' HDU {} has no EXTNAME field'.format(ii))

                if fix:
                    self._fits[ii].header['EXTNAME'] = '{}_{}'.format(self._filename, ii)
                    log.info(' Setting HDU {} EXTNAME field to {}'.format(ii, self._fits[ii].header['EXTNAME']))
                    self._log_text[hdu.name]['data'] = 'Setting HDU {} EXTNAME field to {}\n'.format(ii, self._fits[ii].header['EXTNAME'])

            if hasattr(hdu, 'data') and hdu.data is not None and len(hdu.data.shape) == 3:
                good = True
                self.good_check(good)
                extname = self._fits[ii].header['EXTNAME']

                log.info('  data exists in HDU ({}, {}) and is of shape {}'.format(
                    ii, extname, hdu.data.shape))

                # Check to see if the same size as the others
                if data_shape and not data_shape == hdu.data.shape:
                    log.warning('  Data are of different shapes (previous was {} and this is {})'.format(data_shape,
                                                                                                         hdu.data.shape))

                data_shape = hdu.data.shape

        if not good:
            self.good_check(False)
            log.error('  Can\'t fix lack of data')
            return False

        return good

    def check_ctype1(self, fix=False):
        self._check_ctype(key='CTYPE1', correct='RA---TAN', fix=fix)

    def check_ctype2(self, fix=False):
        self._check_ctype(key='CTYPE2', correct='DEC--TAN', fix=fix)

    def check_ctype3(self, fix=False):
        self._check_ctype(key='CTYPE3', correct='WAVE', fix=fix)

    def check_cunit1(self, fix=False):
        self._check_ctype(key='CUNIT1', correct='deg', fix=fix)

    def check_cunit2(self, fix=False):
        self._check_ctype(key='CUNIT2', correct='deg', fix=fix)

    def check_cunit3(self, fix=False):
        self._check_ctype(key='CUNIT3', correct=self._units_titles, fix=fix)

    def _check_ctype(self, key, correct, fix=False):
        """
        Check the header key and make sure it is the correct value

        :param: fits_file: The open fits file
        :param: fix: boolean whether to fix it or not
        :return: boolean whether it is good or not
        """
        log.debug('In check for {}'.format(key))

        for ii, hdu in enumerate(self._fits):
            if ii == 0 or (hasattr(hdu, 'data') and hdu.data is not None and len(hdu.data.shape) == 3):
                if key not in hdu.header:
                    hdu.header[key] = "NONE"
                    self.good_and_fix(hdu, key, correct, fix, ii)
                else:
                    self.good_and_fix(hdu, key, correct, fix, ii)

    def good_and_fix(self, hdu, key, correct, fix, ii):
        """
        Does as the name implies, checks to see if the hdu.header[key] equals the correct value, if it does not
        and fix is True, the correct value is inserted and passed back to CubeViz
        :param hdu: One of the headers from the original FITS file
        :param key: The header keyword to be checked
        :param correct: The correct value of the header keyword
        :param fix: Whether or not to fix the header to the correct value
        :param ii: The index of the hdu within the FITS file
        :return:
        """
        # (i.e. Angstroms instead of Angstrom will be corrected and added)
        if hdu.header[key] not in correct and len(hdu.header[key]) > 0 and hdu.header[key][:-1] in correct:
            self.good_check(False)
            self._log_text[hdu.name][key] = "{} is {}, setting to {}\n".format(key, hdu.header[key], hdu.header[key][:-1])
            log.info("{} is {}, setting to {} in header[{}]".format(key, hdu.header[key], hdu.header[key][:-1], ii))
            hdu.header[key] = hdu.header[key][:-1]

        elif not hdu.header[key] in correct and fix:
            self.good_check(False)
            if isinstance(correct, list):
                self._log_text[hdu.name][key] = "{} is {}, setting to {}\n".format(key, hdu.header[key], correct[0])
                log.info("{} is {}, setting to {} in header[{}]".format(key, hdu.header[key], correct[0], ii))
                hdu.header[key] = correct[0]
            else:
                self._log_text[hdu.name][key] = "{} is {}, setting to {}\n".format(key, hdu.header[key], correct)
                log.info("{} is {}, setting to {} in header[{}]".format(key, hdu.header[key], correct, ii))
                hdu.header[key] = correct

        elif not hdu.header[key] in correct and not fix:
            self.good_check(False)
            log.info("{} is {}, should equal {} in header[{}]".format(key, hdu.header[key], correct[0], ii))
            self._log_text[hdu.name][key] = "{} is {}, should equal {}".format(key, hdu.header[key], correct[0])

    def get_log_output(self):

        output = self._log_text['>front'] + '\n'

        for log_key, log_value in sorted(self._log_text.items()):
            if log_key == '>front':
                continue

            output += 'HDU {}\n'.format(log_key)

            if isinstance(log_value, str):
                output += '    {}'.format(log_value)
                continue

            for _, log_line in sorted(log_value.items()):
                output += '    {}'.format(log_line)

            output += '\n'

        return output

    def good_check(self, good):
        if good and self._good:
            self._good = True
        if not good:
            self._good = False

    def get_good(self):
        return self._good



# Licensed under a 3-clause BSD style license - see LICENSE.rst

def get_package_data():
    return {
        _ASTROPY_PACKAGE_NAME_ + '.data_factories': ['configurations/*.yaml', 'ifucube_popup.ui']}
